"""Code for the components of the simulation results viewer that generate
the data graph.
"""

import itertools
import math
import random
from typing import Any, Dict, List

# Mypy type checking disabled for packages that are not PEP 561-compliant
import dash                              # type: ignore
import dash_bootstrap_components as dbc  # type: ignore
import plotly.graph_objects as go        # type: ignore

from mahautils.multics.sim_results_viewer.constants import SIM_RESULTS_DICT_T

uirevision_counter = itertools.count(2)


def graph():
    """Creates the main plot where simulation results are displayed"""
    default_height_percent = 90

    return dbc.Spinner(
        dash.dcc.Graph(
            id='plotly-graph',
            figure={},
            style={
                'height': f'{default_height_percent}vh',
            },

            # Styling options
            config={
                'editSelection': True,
                'edits': {
                    'annotationPosition': True,
                    'annotationTail': True,
                    'annotationText': True,
                    'legendPosition': True,
                },
                'modeBarButtonsToAdd': [
                    'drawclosedpath',
                    'drawopenpath',
                    'drawline',
                    'drawrect',
                    'drawcircle',
                    'eraseshape',
                ],
                'toImageButtonOptions': {
                    'format': 'svg',
                    'width': None,
                    'height': None,
                    'scale': 1,
                },
                'scrollZoom': True,
                'showAxisDragHandles': True,
            },
        ),
        delay_show=100,
    )


def random_hex_color() -> str:
    """Generates a hex code for a random color"""
    return '#' + ''.join([f'{random.randint(0, 255):02x}'  # nosec: B311
                          for _ in range(3)])


def update_graph(config_general: dict, config_x: dict, config_y: dict,
                 sim_results_files: SIM_RESULTS_DICT_T, file_metadata: dict):
    """Generates the plot of simulation results based on the current
    user-specified configuration options"""
    figure = go.Figure()
    append_units = bool(config_general['append_units'])

    ## Plot simulation results data ##
    y_axes: List[Dict[str, Any]] = config_y['axes']
    num_active_axes = sum(x['enabled'] for x in y_axes)
    width_per_axis = float(config_general['width_per_y_axis'])

    # Settings for x-axis
    if (axis_title := config_x['axis_title']) not in (None, ''):
        if append_units:
            axis_title += f' [{config_x["units"]}]'

        figure.update_layout(xaxis={'title': axis_title, 'color': 'black'})

    x_domain_min = max(0, width_per_axis*(num_active_axes-1))
    figure.update_layout(
        xaxis={'domain': [x_domain_min, 1]})

    if (dtick := config_x['tick_spacing']) not in (None, ''):
        figure.update_xaxes(dtick=dtick)

    set_xmin = config_x['xmin'] not in (None, '')
    xmin = config_x['xmin'] if set_xmin else math.inf
    set_xmax = config_x['xmax'] not in (None, '')
    xmax = config_x['xmax'] if set_xmax else -math.inf

    if (
        len(y_axes) > 0
        and (config_x['variable'] not in (None, ''))
        and (config_x['units'] not in (None, ''))
    ):
        # Settings for y-axes
        y_axis_data: Dict[str, Any]
        i = 0
        for y_axis_data in y_axes:
            if not y_axis_data['enabled']:
                continue

            plot_data: Dict[str, Any] = {'yaxis': f'y{i+1 if i > 0 else ""}'}

            set_ymin = y_axis_data['ymin'] not in (None, '')
            ymin = y_axis_data['ymin'] if set_ymin else math.inf
            set_ymax = y_axis_data['ymax'] not in (None, '')
            ymax = y_axis_data['ymax'] if set_ymax else -math.inf

            for trace in y_axis_data['traces']:
                if (
                    trace['file'] in (None, '')
                    or trace['variable'] in (None, '')
                    or trace['units'] in (None, '')
                ):
                    # If the file, variable, and units are not all provided,
                    # omit the trace from the plot
                    continue

                if not (
                    trace['enabled']
                    and file_metadata[trace['file']]['enabled']
                ):
                    continue

                try:
                    sim_results = sim_results_files[trace['file']]
                except KeyError as exception:
                    exception.args = (
                        'Unable to find simulation results file named '
                        f'"{exception.args[0]}" -- have you uploaded this file '
                        'on the "Data Files" tab?',)
                    raise

                plot_data['name'] = (
                    f'{trace["name"]}'
                    + (f' [{trace["units"]}]' if append_units else '')
                )

                plot_data['x'] = sim_results.get_data(config_x['variable'],
                                                      config_x['units'])
                plot_data['y'] = sim_results.get_data(trace['variable'],
                                                      trace['units'])
                plot_data['line'] = trace['style']

                if set_xmin != set_xmax:
                    if not set_xmin:
                        xmin = min(xmin, plot_data['x'].min())
                    if not set_xmax:
                        xmax = max(xmax, plot_data['x'].max())

                if set_ymin != set_ymax:
                    if not set_ymin:
                        ymin = min(ymin, plot_data['y'].min())
                    if not set_ymax:
                        ymax = max(ymax, plot_data['y'].max())

                figure.add_trace(go.Scatter(**plot_data))

            tick_options = {}

            if y_axis_data['tick_spacing'] not in (None, ''):
                tick_options['dtick'] = y_axis_data['tick_spacing']

            if set_ymin or set_ymax:
                tick_options['range'] = [ymin, ymax]

            if i == 0:
                figure.update_layout(
                    yaxis={
                        'title': str(y_axis_data['axis_title']),
                        'color': str(y_axis_data.get('color', 'black')),
                        **tick_options,
                    },
                )
            else:
                figure.update_layout(**{
                    f'yaxis{i+1}': {
                        'title': str(y_axis_data['axis_title']),
                        'anchor': 'free',
                        'overlaying': 'y',
                        'side': 'left',
                        'position': (width_per_axis * (num_active_axes - i - 1)),
                        'color': str(y_axis_data.get('color', 'black')),
                        **tick_options,
                    }
                })

            i += 1

    if set_xmin or set_xmax:
        figure.update_xaxes(range=[xmin, xmax])

    ## Plot formatting settings ##
    figure.update_layout(margin={'t': 7.5, 'r': 10}, showlegend=True)

    default_axes_settings = {
        # Border around plot
        'showline': True, 'linewidth': 1, 'linecolor': 'black', 'mirror': True,

        # Settings for x=0 and y=0 axes
        'zeroline': True, 'zerolinewidth': 1, 'zerolinecolor': 'black',

        # Settings for gridlines
        'showgrid': True, 'gridwidth': 1, 'gridcolor': '#d2d2d2',
    }
    figure.update_xaxes(**default_axes_settings)
    figure.update_yaxes(**default_axes_settings)

    # Remove y=0 zero line (can be easily confused with simulation results data)
    figure.update_yaxes(zeroline=False)

    figure.update_layout(
        plot_bgcolor=config_general['background_color'])

    # Plot title
    if config_general['title'] not in (None, ''):
        figure.update_layout(
            title={
                'text': f'<b>{config_general["title"]}</b>',
                'x': 0.5,
                'xanchor': 'center',
            },
            title_font_color='black',
            margin={'t': 40},
        )

    # Gridlines
    figure.update_xaxes(showgrid=bool(config_general['grid_x']))
    figure.update_yaxes(showgrid=bool(config_general['grid_y']))

    # Cursor hover settings
    figure.update_layout(
        hovermode=config_general['hovermode'],
        hoverlabel={'namelength': -1},  # prevents truncating hover text
    )

    # Whether to change settings like zoom and which legend entries are active
    # when figure is updated by Dash
    if config_general['freeze_uirevision']:
        # For the plot UI state not to be updated, "uirevision" must be set to
        # a constant, nonzero value (plotly/graphing-library-docs#116)
        figure.update_layout(uirevision=1)
    else:
        figure.update_layout(uirevision=next(uirevision_counter))

    return figure



"""Constants shared throughout the MahaUtils SimViewer GUI.
"""

# Importing the package version is a cylic import, but because the
# `__version__` variable is defined at the beginning of
# `mahautils/__init__.py`, it is cached and cyclic import does not
# cause problems
from mahautils import __version__ as VERSION  # pylint: disable=R0401  # noqa: F401

from mahautils.multics.simresults import SimResults
from mahautils.utils.dictionary import Dictionary


# GENERAL PROJECT SETTINGS ---------------------------------------------------
PROJECT_NAME = 'MahaUtils'
GUI_LONG_NAME = 'Simulation Results Viewer'
GUI_SHORT_NAME = 'SimViewer'

REPO_URL = 'https://github.com/nathan-hess/maha-research-utils'


# TYPE ALIASES ---------------------------------------------------------------
# Type of global variable which stores simulation results files
SIM_RESULTS_DICT_T = Dictionary[str, SimResults]


# STYLING --------------------------------------------------------------------
# Margin between tab bar and content in configuration panel
TAB_BAR_PADDING = '20px'

# Styling for color picker UI elements
COLOR_PICKER_STYLE = {'width': 75, 'height': 50}

# Margin below HTML <p> elements used to describe features to users
UI_DESCRIPTION_MARGIN_BELOW = '5px'

# Option height in dropdowns when selecting simulation results variables
# to plot (https://community.plotly.com/t/long-dropdown-values-overlap/14843)
SIM_VAR_DROPDOWN_OPTION_HEIGHT = 48



"""Data structures for storing program data in the browser.
"""

from typing import Any, Dict

# Mypy type checking disabled for packages that are not PEP 561-compliant
import dash  # type: ignore


def file_metadata_store():
    """Creates browser session storage for metadata about uploaded simulation
    results files

    Format
    ------
    ```
    {
        'sim_results_file_name_1': {
            'enabled': True,
        },
        'sim_results_file_name_2': {
            'enabled': False,
        },
    }
    ```
    """
    return dash.dcc.Store(
        id='data-file-store',
        storage_type='session',
    )


default_plot_config_general: Dict[str, Any] = {
    'title': None,
    'background_color': '#ffffff',
    'grid_x': True,
    'grid_y': True,
    'append_units': False,
    'width_per_y_axis': 0.1,
    'hovermode': 'closest',
    'freeze_uirevision': False,
    'image_export_type': 'png',
    'image_export_width': None,
    'image_export_height': None,
    'image_export_scale': 1,
}

default_plot_config_x: Dict[str, Any] = {
    'axis_title': None,
    'variable': None,
    'units': None,
    'xmin': None,
    'xmax': None,
    'tick_spacing': None,
}

default_plot_config_y: Dict[str, Any] = {
    'axes': [],
}

default_y_axis_settings: Dict[str, Any] = {
    'axis_title': None,
    'enabled': True,
    'color': '#000000',
    'ymin': None,
    'ymax': None,
    'tick_spacing': None,
    'traces': [],
}

default_trace_settings: Dict[str, Any] = {
    'name': None,
    'enabled': True,
    'file': None,
    'variable': None,
    'units': None,
    'style': {
        'color': '#1f77b4',
        'width': 2,
        'dash': 'solid',
    },
}


def plot_config_general_store():
    """Creates browser session storage for general plot settings
    """
    return dash.dcc.Store(
        id='plot-config-general-store',
        storage_type='session',
        data=default_plot_config_general,
    )


def plot_config_x_store():
    """Creates browser session storage for x-axis plot settings
    """
    return dash.dcc.Store(
        id='plot-config-x-store',
        storage_type='session',
        data=default_plot_config_x,
    )


def plot_config_y_store():
    """Creates browser session storage for y-axes plot settings
    """
    return dash.dcc.Store(
        id='plot-config-y-store',
        storage_type='session',
        data=default_plot_config_y,
    )



"""Information box showing software version and credits.
"""

# Mypy type checking disabled for packages that are not PEP 561-compliant
import dash                              # type: ignore
import dash_bootstrap_components as dbc  # type: ignore
import numpy as np
import plotly                            # type: ignore
import pyxx

from .constants import (
    GUI_LONG_NAME,
    GUI_SHORT_NAME,
    PROJECT_NAME,
    REPO_URL,
    VERSION,
)


def info_box():
    """Creates an "about" box showing program version and acknowledgements"""
    contents = [
        dbc.ModalHeader(
            dbc.ModalTitle(f'{PROJECT_NAME} {GUI_SHORT_NAME} v{VERSION}'),
            close_button=True,
        ),
        dbc.ModalBody([
            dash.html.P(
                [
                    (f'{PROJECT_NAME} is an open source project, and all '
                     'source code is available through '),
                    dash.html.A(
                        ['GitHub ', dash.html.I(className='bi bi-github')],
                        href=REPO_URL,
                        target='_blank',
                    ),
                    (f'.  The {PROJECT_NAME} {GUI_LONG_NAME} ({GUI_SHORT_NAME}) '
                     'graphical user interface is powered by '),
                    dash.html.A(
                        'Dash',
                        href='https://plotly.com/dash/',
                        target='_blank',
                    ),
                    '.',
                ]
            ),
            dash.html.Br(),
            dash.html.P('Dependency versions:'),
            dash.html.Ul([
                dash.html.Li(f'Dash: v{dash.__version__}'),
                dash.html.Li(f'Dash Bootstrap Components: v{dbc.__version__}'),
                dash.html.Li(f'NumPy: v{np.__version__}'),
                dash.html.Li(f'Plotly: v{plotly.__version__}'),
                dash.html.Li(f'PyXX: v{pyxx.__version__}'),
            ]),
        ]),
        dbc.ModalFooter(
            dbc.Button(
                'Close',
                id='info-box-close-button',
            ),
        ),
    ]

    return dbc.Modal(
        contents,
        id='info-button-modal',
        scrollable=True,
        size='lg',
        is_open=False,
    )



"""**Utilities to support other code in the MahaUtils package**

The :py:mod:`mahautils.utils` module provides utility functions and other
general-purpose code applicable to a variety of MahaUtils modules.
"""

from .arrays import to_np_1D_array
from .capture_printing import CaptureStderr
from .dictionary import Dictionary



"""Panel tab for configuring y-axis plot settings.
"""

# Mypy type checking disabled for packages that are not PEP 561-compliant
import dash                              # type: ignore
import dash_bootstrap_components as dbc  # type: ignore

from mahautils.multics.sim_results_viewer.constants import (
    SIM_RESULTS_DICT_T,
    SIM_VAR_DROPDOWN_OPTION_HEIGHT,
    TAB_BAR_PADDING,
    UI_DESCRIPTION_MARGIN_BELOW,
)
from mahautils.multics.sim_results_viewer.store import (
    default_plot_config_y,
    default_trace_settings,
    default_y_axis_settings,
)
from mahautils.multics.sim_results_viewer.ui_elements import (
    color_picker,
    numbered_item_selector,
    tick_range_step,
)


def plot_settings_y():
    """Creates the configuration panel tab with settings for the y-axes"""
    return [
        dash.html.Div(
            render_y_settings(default_plot_config_y, {}, {}, 0, 0),
            id='plot-config-y-settings',
            style={'marginTop': TAB_BAR_PADDING},
        ),
    ]


def render_y_settings(config_y: dict, sim_results_files: SIM_RESULTS_DICT_T,
                      file_metadata: dict,
                      selected_axis: int, selected_data_series: int):
    """Generates the UI elements that allow the user to configure y-axis
    plot settings"""
    # Extract axis data
    num_axes = len(config_y['axes'])
    is_axis_selected = (num_axes != 0)

    axis = config_y['axes'][selected_axis] if num_axes > 0 \
        else default_y_axis_settings

    # Extract data series/trace data
    num_traces = len(axis['traces'])
    is_trace_selected = (num_traces != 0)

    trace = axis['traces'][selected_data_series] if num_traces > 0 \
        else default_trace_settings

    file_selected = trace['file'] is not None
    file_enabled = ((trace['file'] in file_metadata)
                    and file_metadata[trace['file']]['enabled'])

    # Get list of possible simulation results variables to plot
    if trace['file'] in (None, ''):
        variable_options = []
    else:
        variable_options = [{
            'label': (
                f'{var} — '
                f'{sim_results_files[trace["file"]].get_description(var)}'
            ),
            'value': var,
        } for var in sim_results_files[trace['file']].variables]

    return [
        # Axis selector and buttons for creating and deleting axes
        dash.html.H5('Y-Axis Selector', style={'marginBottom': '10px'}),
        numbered_item_selector(
            pagination_id='y-axis-selector',
            add_button_id='add-y-axis-button',
            hide_show_button_id='hide-show-y-axis-button',
            delete_button_id='delete-y-axis-button',
            name='y-axis',
            name_plural='y-axes',
            num_items=num_axes,
            active_page=selected_axis+1,
            is_active_page_shown=axis['enabled'],
        ),

        # Selected axis properties
        dash.html.Div([
            # Axis formatting options
            dash.html.H4('Formatting'),
            dash.html.Hr(),
            dash.html.H5('Axis Title'),
            dbc.Input(
                debounce=True,
                id={'component': 'plot-config', 'tab': 'y',
                    'field': 'title'},
                placeholder='Enter a title for the y-axis...',
                value=axis['axis_title'],
            ),
            dash.html.Br(),
            dash.html.H5('Axis Color'),
            color_picker(
                identifier={'component': 'plot-config', 'tab': 'y',
                            'field': 'axis-color'},
                current_color=axis['color'],
            ),
            dash.html.Br(),
            tick_range_step(
                range_min_id={'component': 'plot-config', 'tab': 'y',
                              'field': 'ymin'},
                range_max_id={'component': 'plot-config', 'tab': 'y',
                              'field': 'ymax'},
                tick_spacing_id={'component': 'plot-config', 'tab': 'y',
                                 'field': 'tick_spacing'},
                range_min_val=axis['ymin'],
                range_max_val=axis['ymax'],
                tick_spacing_val=axis['tick_spacing'],
            ),
            dash.html.Br(),
            dash.html.Br(),

            # Axis data/traces selection
            dash.html.H4('Data Series Configuration'),
            dash.html.Hr(),
            numbered_item_selector(
                pagination_id='y-data-selector',
                add_button_id='add-y-data-button',
                hide_show_button_id='hide-show-y-data-button',
                delete_button_id='delete-y-data-button',
                name='data series',
                name_plural='data series',
                num_items=num_traces,
                active_page=selected_data_series+1,
                is_active_page_shown=trace['enabled'],
                disable_hide_show_button=(not file_enabled),
            ),
            dash.html.Div([
                dash.html.Div(
                    [
                        dash.html.P(
                            'The show/hide controls for this data series are '
                            'disabled because the file has been disabled on the '
                            '"Data Files" tab.  You can edit the settings '
                            'below, but the data series will not be displayed '
                            'in the plot until you enable the file.',
                            style={'fontStyle': 'italic'},
                        ),
                        dash.html.Br(),
                    ],
                    hidden=(file_enabled or not file_selected),
                ),
                dash.html.H5('Legend Title'),
                dash.html.P(
                    'Enter a legend title for the data series. Use "<br>" to '
                    'split long lines',
                    style={'marginBottom': UI_DESCRIPTION_MARGIN_BELOW},
                ),
                dbc.Input(
                    debounce=True,
                    id={'component': 'plot-config', 'tab': 'y',
                        'field': 'legend-title'},
                    value=trace['name'],
                ),
                dash.html.Br(),
                dash.html.H5('Simulation Results File'),
                dash.dcc.Dropdown(
                    options=list(sim_results_files.keys()),
                    value=trace['file'],
                    id={'component': 'plot-config', 'tab': 'y',
                        'field': 'trace-file'},
                    clearable=True,
                    multi=False,
                ),
                dash.html.Br(),
                dash.html.H5('Simulation Results Variable'),
                dash.html.Div(dbc.Row([
                    dbc.Col(
                        dash.html.H6('Variable'),
                        width=2,
                    ),
                    dbc.Col(
                        dash.dcc.Dropdown(
                            options=variable_options,
                            value=trace['variable'],
                            id={'component': 'plot-config', 'tab': 'y',
                                'field': 'trace-variable'},
                            clearable=True,
                            disabled=(trace['file'] in (None, '')),
                            multi=False,
                            optionHeight=SIM_VAR_DROPDOWN_OPTION_HEIGHT,
                        ),
                        width=10,
                    ),
                ]), style={'marginBottom': UI_DESCRIPTION_MARGIN_BELOW}),
                dbc.Row([
                    dbc.Col(
                        dash.html.H6('Units'),
                        width=2,
                    ),
                    dbc.Col(
                        dbc.Input(
                            debounce=True,
                            value=trace['units'],
                            id={'component': 'plot-config', 'tab': 'y',
                                'field': 'units'},
                            disabled=(trace['file'] in (None, '')
                                      or trace['variable'] in (None, '')),
                            style={'width': 150},
                        ),
                        width=10,
                    ),
                ]),
                dbc.Alert(
                    [
                        dash.html.H5([
                            dash.html.I(className='fa fa-exclamation-triangle'),
                            ' WARNING',
                        ]),
                        dash.html.Hr(),
                        dash.html.P('The plot will not be updated until you '
                                    'have selected a simulation results file, '
                                    'variable, AND units for the data series'),
                    ],
                    color='danger',
                    style={
                        'width': '95%',
                        'marginTop': '10px',
                        'marginLeft': '10px',
                        'marginRight': '10px',
                    },
                    is_open=(
                        trace['file'] in (None, '')
                        or trace['variable'] in (None, '')
                        or trace['units'] in (None, '')
                    ),
                    dismissable=False,
                ),
                dash.html.Br(),
                dash.html.H5('Line Color'),
                color_picker(
                    identifier={'component': 'plot-config', 'tab': 'y',
                                'field': 'trace-color'},
                    current_color=trace['style']['color'],
                ),
                dash.html.Br(),
                dash.html.H5('Line Width'),
                dbc.Input(
                    debounce=True,
                    id={'component': 'plot-config', 'tab': 'y',
                        'field': 'trace-width'},
                    value=trace['style']['width'],
                ),
                dash.html.Br(),
                dash.html.H5('Line Style'),
                dash.dcc.Dropdown(
                    options=['solid', 'dot', 'dash', 'longdash',
                             'dashdot', 'longdashdot'],
                    value=trace['style']['dash'],
                    id={'component': 'plot-config', 'tab': 'y',
                        'field': 'trace-dash'},
                    clearable=False,
                    multi=False,
                ),
            ], hidden=(not is_trace_selected)),
        ], hidden=(not is_axis_selected))
    ]



"""The classes in this module are intended to store a group of related
:py:class:`Layer` objects.
"""

import itertools
from typing import Optional, Union

# Mypy type checking disabled for packages that are not PEP 561-compliant
import plotly.graph_objects as go  # type: ignore
import pyxx

from mahautils.shapes.geometry.point import Array_Float2
from mahautils.shapes.geometry.point2D import CartesianPoint2D
from .layer import Layer
from .plotting import _figure_config, _create_blank_plotly_figure


class Canvas(pyxx.arrays.TypedListWithID[Layer]):
    """An object for storing a set of layers

    A :py:class:`Canvas` is intended to store a set of related layers, allowing
    more complex, multi-shape geometries to be constructed.  For instance, one
    potential application would be to use two :py:class:`Layer` objects to
    store the disjoint polygons corresponding to the high-pressure ports and
    the low-pressure ports in an axial piston pump, and then to group these
    layers at each instant in time in a :py:class:`Canvas`.  Then, multiple
    :py:class:`Canvas` objects could be used to store layers for every time
    step in the simulation.

    The :py:class:`Canvas` inherits from Python's ``MutableSequence`` type, so
    most operations that can be performed for a Python ``list`` (append, pop,
    insert, etc.) are valid operations for a :py:class:`Canvas`.
    """

    _id = itertools.count(0)

    def __init__(self, *layers: Layer, name: Optional[str] = None,
                 print_multiline: bool = True) -> None:
        """Creates a new canvas to store layers

        Creates a new :py:class:`Canvas` object in which a set of
        :py:class:`Layer` objects can be stored.

        Parameters
        ----------
        *layers : Layer, optional
            Layers to add to the canvas after the canvas is created (default
            is to add no layers)
        name : str, optional
            A descriptive name to identify the canvas.  If not provided, the
            canvas name is generated automatically
        print_multiline : bool, optional
            Whether to return a printable string representation of the list in
            multiline format (default is ``True``).  Multiline format places
            each item in the list on its own line
        """
        super().__init__(
            *layers,
            list_type=Layer,
            print_multiline=print_multiline,
            multiline_padding=1
        )

        # Set canvas name
        self.name = f'canvas{self.id}' if name is None else name

    @property
    def name(self) -> str:
        """A descriptive name to identify the canvas"""
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = str(name)

    @property
    def num_layers(self) -> int:
        """The number of layers in the canvas"""
        return len(self)

    def plot(self, units: Optional[str] = None,
             figure: Optional[go.Figure] = None,
             show: bool = True, return_fig: bool = False,
             ) -> Union[go.Figure, None]:
        """Plots the shapes in the canvas

        Creates a Plotly figure illustrating the shapes in all layers of the
        canvas, or optionally appends traces for each shape to an existing
        figure.  The figure can be opened in a browser (default behavior)
        and/or returned (to allow subsequent user-specific customizations).

        Parameters
        ----------
        units : str, optional
            If units are provided, it will be verified that all plotted shapes
            have these units in their :py:attr:`Shape2D.units` attribute
            (default is ``None``, which performs no unit checks).  Additionally,
            if the ``figure`` argument was not provided, the specified units
            will be included in the axis titles
        figure : go.Figure, optional
            A Plotly figure.  If provided, rather than creating a new figure
            from scratch, the layer's shapes will be added as new traces in
            the provided figure (default is ``None``, which creates a new
            figure from scratch)
        show : bool, optional
            Whether to open the figure in a browser (default is ``True``)
        return_fig : bool, optional
            Whether to return the figure (default is ``False``)

        Returns
        -------
        go.Figure
            A Plotly figure depicting the canvas.  Returned if and only if
            ``return_fig`` is ``True``
        """
        if not isinstance(figure, go.Figure):
            figure = _create_blank_plotly_figure(units)

        for layer in self:
            figure = layer.plot(units=units, figure=figure,
                                show=False, return_fig=True)

        if show:
            figure.show(config=_figure_config)

        if return_fig:
            return figure

        return None

    def rotate(self, center: Union[Array_Float2, CartesianPoint2D],
               angle: float, angle_units: str = 'rad') -> None:
        """Rotates all shapes in all layers of the canvas a given angle in
        the :math:`xy`-plane about a user-specified point

        Parameters
        ----------
        center : list or tuple or CartesianPoint2D
            The center of rotation about which to rotate the shapes
        angle : float
            The angle by which to rotate the shapes about ``center``
        angle_units : str, optional
            The units (radians or degrees) of the ``angle`` argument.  Must be
            either ``'rad'`` or ``'deg'`` (default is ``'rad'``)
        """
        for layer in self:
            layer.rotate(center=center, angle=angle, angle_units=angle_units)

    def translate(self, x: float = 0, y: float = 0) -> None:
        """Translates all shapes in all layers of the canvas a user-specified
        distance in the :math:`x`- and/or :math:`y`-directions

        Parameters
        ----------
        x : float, optional
            The distance to translate the shapes in the :math:`x`-direction
            (default is ``0``)
        y : float, optional
            The distance to translate the shapes in the :math:`y`-direction
            (default is ``0``)
        """
        for layer in self:
            layer.translate(x=x, y=y)



"""Function for generating pop-up error boxes"""

from typing import Dict

# Mypy type checking disabled for packages that are not PEP 561-compliant
import dash                              # type: ignore
import dash_bootstrap_components as dbc  # type: ignore


def error_box(identifier: Dict[str, str]):
    """Creates a pop-up box for displaying error messages"""
    text_id = {k: (v + '-text') for k, v in identifier.items()}

    contents = [
        dbc.ModalHeader(
            dbc.ModalTitle(
                dash.html.H4([
                    dash.html.I(className='fa fa-exclamation-triangle'),
                    ' Error',
                ], style={'color': 'red'})),
            close_button=True,
        ),
        dbc.ModalBody(dash.html.Div(id=text_id)),
    ]

    return dbc.Modal(
        contents,
        id=identifier,
        scrollable=True,
        size='lg',
        is_open=False,
    )


def generate_error_box_text(message, exception: Exception):
    """Generates error box text, beginning with a user-specified error message
    followed by the Python exception message"""
    return [
        message,
        dash.html.Br(),
        dash.html.P('Debugging Details', style={'fontWeight': 'bold'}),
        dash.html.P('Here is the exception that triggered this error:'),
        dash.html.P(f'{exception.__class__.__name__}: {exception}',
                    style={'fontFamily': 'monospace'})
    ]



"""Class for representing arbitrary geometry
"""

from typing import Optional, Union


class Geometry:
    """Base class for representing arbitrary geometry
    """

    def __init__(self, units: Optional[str] = None) -> None:
        """Creates a new object representing arbitrary geometry

        Parameters
        ----------
        units : str, optional
            The units in which the geometry is defined, or ``None`` to
            indicate dimensionless geometry or that units are to be ignored
            (default is ``None``)
        """
        self.units = units

    @property
    def units(self) -> Union[str, None]:
        """The units in which the geometry is defined"""
        return self._units

    @units.setter
    def units(self, units: Optional[str]) -> None:
        if units is None:
            self._units = None
        else:
            self._units = str(units)

    def _has_identical_units(self, geometry: 'Geometry') -> bool:
        """Checks that the units of two :py:class:`Geometry` objects are
        identical

        Parameters
        ----------
        geometry : Geometry
            Another :py:class:`Geometry` instance or subclass whose units to
            check

        Returns
        -------
        bool
            Returns ``True`` if both objects have units of ``None`` or
            identical units, and ``False`` otherwise
        """
        if self.units is None:
            return geometry.units is None

        if geometry.units is None:
            return self.units is None

        return self.units == geometry.units



"""Code that generates the table of loaded simulation results files.
"""

from typing import Any, Dict

# Mypy type checking disabled for packages that are not PEP 561-compliant
import dash                              # type: ignore
import dash_bootstrap_components as dbc  # type: ignore

from mahautils.multics.sim_results_viewer.constants import SIM_RESULTS_DICT_T


def empty_file_table():
    """Generates an empty table which can display a list of currently uploaded
    simulation results files"""
    return dbc.Table(
        [
            dash.html.Thead(dash.html.Tr(_table_header())),
            dash.html.Tbody(id='file-list-table-body'),
        ],
    )


def generate_file_table_body(sim_results_files: SIM_RESULTS_DICT_T,
                             metadata: Dict[str, Dict[str, Any]]):
    """Populates the simulation results file table with all currently uploaded
    simulation results files"""
    description_style = {
        'fontSize': 12.5,
        'marginTop': 0,
        'marginBottom': 0,
        'paddingTop': 0,
        'paddingBottom': 0,
    }

    description_header_style = description_style.copy()
    description_header_style['fontSize'] += 1.5

    contents = []
    for _, key in enumerate(sim_results_files.keys()):
        display_metadata = []

        title = sim_results_files[key].title
        sim_version = sim_results_files[key].sim_version
        if (title is not None) or (sim_version is not None):
            display_metadata.append(
                dash.html.B('Simulation Metadata',
                            style=description_header_style)
            )

            if title is not None:
                display_metadata.append(
                    dash.html.P(
                        [dash.html.B('Title: '), title],
                        style=description_style,
                    )
                )

            if sim_version is not None:
                display_metadata.append(
                    dash.html.P(
                        [dash.html.B('Version: '), sim_version],
                        style=description_style,
                    )
                )

        maha_multics_version = sim_results_files[key].maha_multics_version
        maha_multics_commit = sim_results_files[key].maha_multics_commit
        if (maha_multics_version is not None) or (maha_multics_commit is not None):
            if (title is not None) or (sim_version is not None):
                display_metadata.append(dash.html.P())

            display_metadata.append(
                dash.html.B('Maha Multics Metadata',
                            style=description_header_style)
            )

            if maha_multics_version is not None:
                display_metadata.append(
                    dash.html.P(
                        [dash.html.B('Version: '),
                         maha_multics_version],
                        style=description_style,
                    )
                )

            if maha_multics_commit is not None:
                display_metadata.append(
                    dash.html.P(
                        [dash.html.B('Commit: '),
                         maha_multics_commit],
                        style=description_style,
                    )
                )

        if len(display_metadata) == 0:
            file_id_description = dash.html.H6(key)
        else:
            file_id_description = dash.html.Details(
                [dash.html.Summary(key, style={'marginBottom': 5})]
                + display_metadata
            )

        contents.append(dash.html.Tr([
            dash.html.Td(dbc.Switch(
                value=metadata[key]['enabled'],
                id={'component': 'file-table-switch', 'key': key},
            )),
            dash.html.Td(file_id_description),
            dash.html.Td(dbc.Button(
                dash.html.I(className='bi bi-trash'),
                color='danger',
                id={'component': 'file-table-delete-button', 'key': key},
            )),
        ]))

    return contents


def _table_header():
    return [
        dash.html.Th('Enabled', style={'width': '15%'}),
        dash.html.Th('Name', style={'width': '70%'}),
        dash.html.Th('Delete', style={'width': '15%'}),
    ]



"""Customized exceptions for the :py:mod:`mahautils.multics` module
"""

# GENERAL EXCEPTIONS
class MahaMulticsFileFormatError(Exception):
    """General exception for errors related to Maha Multics file parsers"""

class FileNotParsedError(MahaMulticsFileFormatError):
    """Error thrown when attempting to retrieve information from a file, but
    before the file has been read or contents populated"""


# FLUID PROPERTY FILES
class FluidPropertyFileError(MahaMulticsFileFormatError):
    """General exception for errors with fluid property files"""


# SIMULATION RESULTS FILES
class SimResultsError(MahaMulticsFileFormatError):
    """General exception for errors related to Maha Multics simulation results
    files"""

class InvalidSimResultsFormatError(SimResultsError):
    """Maha Multics simulation results file content does not match expected
    format"""

class SimResultsKeyError(SimResultsError, KeyError):
    """Attempting to access a given variable in a simulation results file, but
    the variable is not present in the file"""

class SimResultsDataNotFoundError(SimResultsError, ValueError):
    """Error thrown if attempting to access data in a simulation results file
    that has not yet been defined"""

class SimResultsOverwriteError(SimResultsError):
    """Error thrown if overwriting data in a simulation results file that
    should not be overwritten"""


# VTK FILES
class VTKFormatError(MahaMulticsFileFormatError):
    """Error thrown if VTK file contains data in an unexpected format"""

class VTKIdentifierNameError(VTKFormatError, ValueError):
    """Error thrown if VTK file data identifier name has an invalid format"""

class VTKInvalidIdentifierError(IndexError):
    """Error thrown if attempting to retrieve data from a VTK file, but the
    identifier provided by the user does not correspond to a single column
    in the VTK data DataFrame"""


# POLYGON FILES
class PolygonFileError(MahaMulticsFileFormatError):
    """General class for errors involving Maha Multics polygon files"""

class PolygonFileFormatError(PolygonFileError):
    """Error thrown if formatting issues are encountered in a polygon file"""

class PolygonFileMissingDataError(PolygonFileError):
    """Error thrown if attempting to edit data in a polygon file without first
    having defined required attributes"""



"""A web app-like graphical user interface for viewing Maha Multics simulation
results.
"""

# Some linter settings are disabled due to incompatibility with standard Dash
# coding practice or to improve the app's user experience
#
# pylint: disable=unused-argument,broad-exception-caught

import argparse
import copy
import os
import sys
from typing import Any, Dict, List, Optional
import webbrowser

# Mypy type checking disabled for packages that are not PEP 561-compliant
import dash                              # type: ignore
from dash import Input, Output, State    # type: ignore
import dash_bootstrap_components as dbc  # type: ignore
import plotly.graph_objects as go        # type: ignore

from mahautils.utils import Dictionary
from .constants import (
    PROJECT_NAME,
    GUI_SHORT_NAME,
    SIM_RESULTS_DICT_T,
    VERSION,
)
from .error_box import error_box, generate_error_box_text
from .header import app_header
from .info import info_box
from .load_files import (
    load_plot_config,
    load_plot_config_error_message,
    load_simresults,
    plot_config_to_str,
)
from .panel import (
    generate_file_table_body,
    render_general_settings,
    render_x_settings,
    render_y_settings,
    simviewer_config_panel,
)
from .plotting import graph, random_hex_color, update_graph
from .store import (
    default_trace_settings,
    default_y_axis_settings,
    file_metadata_store,
    plot_config_general_store,
    plot_config_x_store,
    plot_config_y_store,
)


## MAIN APP CONFIGURATION ----------------------------------------------------
app = dash.Dash(
    external_stylesheets=[
        dbc.themes.BOOTSTRAP,
        dbc.icons.BOOTSTRAP,
        dbc.icons.FONT_AWESOME,
    ],
    update_title=None,
)
app.title = f'{PROJECT_NAME} {GUI_SHORT_NAME}'


def main(argv: Optional[List[str]] = None) -> int:
    """Main entrypoint for running MahaUtils simulation results viewer"""
    # Command-line argument parsing
    if argv is None:
        argv = sys.argv[1:]

    parser = argparse.ArgumentParser(
        prog='SimViewer',
        description=('The MahaUtils Simulation Results Viewer is a graphical '
                     'tool that aids in reading Maha Multics simulation '
                     'results files and viewing results. Project documentation '
                     'can be found at https://mahautils.readthedocs.io.'),
    )
    parser.add_argument('--port', action='store', type=int, default=8050,
                        help='port on which to serve the app (default is 8050)')
    parser.add_argument('--debug', action='store_true',
                        help=('enable Dash\'s debug mode (more information: '
                              'https://dash.plotly.com/devtools)'))
    parser.add_argument('--no-browser', action='store_true',
                        help=('prevent automatically opening a browser window '
                              f'when launching {GUI_SHORT_NAME}'))
    parser.add_argument('--version', action='version', version=VERSION)

    args = parser.parse_args(argv)

    port = int(args.port)

    # Create and load Dash app
    app.layout = dash.html.Div([
        file_metadata_store(),
        plot_config_general_store(),
        plot_config_x_store(),
        plot_config_y_store(),
        app_header(),
        graph(),
        simviewer_config_panel(),
        info_box(),
        error_box({'component': 'error-box', 'id': 'load-simresults'}),
        error_box({'component': 'error-box', 'id': 'update-config'}),
        error_box({'component': 'error-box', 'id': 'update-plot'}),
    ])

    # Launch browser to display app
    if not bool(args.no_browser) and ('WERKZEUG_RUN_MAIN' not in os.environ):
        # Checking for the "WERKZEUG_RUN_MAIN" variable prevents opening a new
        # browser window each time source code changes are saved in debug mode
        webbrowser.open(f'http://127.0.0.1:{port}/')

    # Run web app
    app.run_server(debug=bool(args.debug), port=port)

    return 0


## GLOBAL VARIABLES FOR DATA STORAGE -----------------------------------------
#
# In general, it is not a good idea to use global variables in Dash apps
# (https://dash.plotly.com/sharing-data-between-callbacks).  However, there
# are several reasons global variables are suitable for this use case:
#
# 1. This app is intended to be run locally, not on a shared server.
# 2. The app needs to be able to load arbitrarily large simulation results
#    files.  Some can be hundreds of megabytes, which will likely exceed
#    many browser quotas.
# 3. The `SimResults` object performs a number of file-parsing and analysis
#    tasks.  If serializing session data to JSON, it would be necessary to
#    perform these tasks on each load, which could be time-consuming for
#    large files and result in poor visual performance.
# 4. Users may want to open the same simulation results files in multiple
#    tabs/windows.  By using a global variable, the same files will be
#    displayed in all browser tabs/windows.

_sim_results_files: SIM_RESULTS_DICT_T = Dictionary()


## DATA FILE MANAGEMENT ------------------------------------------------------
@app.callback(
    Output('file-list-table-body', 'children'),
    Output('data-file-store', 'data'),
    Output('upload-data', 'contents'),
    Output({'component': 'error-box', 'id': 'load-simresults'}, 'is_open'),
    Output({'component': 'error-box-text', 'id': 'load-simresults-text'},
           'children'),
    Input('load-file-button', 'n_clicks'),
    Input({'component': 'file-table-switch', 'key': dash.ALL}, 'value'),
    Input({'component': 'file-table-delete-button', 'key': dash.ALL}, 'n_clicks'),
    State({'component': 'file-table-switch', 'key': dash.ALL}, 'id'),
    State('user-file-name', 'value'),
    State('file-list-table-body', 'children'),
    State('data-file-store', 'data'),
    State('upload-data', 'contents'),
)
def data_file_manager(
    ## Inputs ##
    n_clicks: int,
    enabled_switch_values: Optional[List[bool]],
    delete_button_clicks: Optional[List[int]],
    ## States ##
    enabled_switch_ids: Optional[List[Dict[str, Any]]],
    name: Optional[str],
    current_file_table: Optional[list],
    metadata: Optional[Dict[str, Dict[str, Any]]],
    contents: Optional[str],
):
    """Reads uploaded simulation results from a file"""
    # Ensure that keys in file metadata store match current list of simulation
    # results files.  This is important if, for instance, multiple tabs are
    # opened, as each will have its own session storage.  This step ensures all
    # stores are synced with the global file list
    if metadata is None:
        metadata = {}

    metadata = {
        key: metadata.get(key, {'enabled':  True})
        for key in _sim_results_files.keys()
    }

    # Upload simulation results file or modify stored metadata
    updated_contents = None
    if dash.ctx.triggered_id is None:
        # Callback was triggered by initial page load, so skip
        # other actions and render table
        pass

    elif dash.ctx.triggered_id == 'load-file-button':
        if (contents is not None) and (name is not None):
            try:
                _sim_results_files[name] = load_simresults(contents)
            except Exception as exception:
                error_box_text = generate_error_box_text(
                    dash.html.P(
                        f'Unable to read simulation results file "{name}."'),
                    exception,
                )
                return (current_file_table, metadata, contents,
                        True, error_box_text)
            metadata[name] = {'enabled': True}

    elif dash.ctx.triggered_id['component'] == 'file-table-switch':
        for i, value in enumerate(enabled_switch_values):  # type: ignore
            key = enabled_switch_ids[i]['key']  # type: ignore
            metadata[key]['enabled'] = value

        updated_contents = contents

    elif dash.ctx.triggered_id['component'] == 'file-table-delete-button':
        key = dash.ctx.triggered_id['key']
        del _sim_results_files[key]
        del metadata[key]

        updated_contents = contents

    return (
        generate_file_table_body(_sim_results_files, metadata),
        metadata,
        updated_contents,
        False,
        None,
    )


@app.callback(
    Output('div-user-file-name', 'hidden'),
    Output('user-file-name', 'value'),
    Input('upload-data', 'contents'),
    State('upload-data', 'filename'),
)
def hide_show_upload_filename(contents: Optional[str], filename: Optional[str]):
    """Hides/shows the file naming box and populates the input area with
    a default name based on the uploaded file"""
    if (contents is not None) and (filename is not None):
        if filename.endswith('.txt'):
            filename = filename[:-4]

        return False, filename

    return True, 'NONE'


@app.callback(
    Output('file-overwrite-alert', 'is_open'),
    Output('load-file-button', 'disabled'),
    Output('upload-data', 'filename'),
    Input('user-file-name', 'value'),
    prevent_initial_call=True,
)
def validate_upload_file_name(name: Optional[str]):
    """Checks that the user provided a valid file description for a simulation
    results file"""
    overwrite_alert_open = False
    load_button_disabled = False

    if (name is None) or (name in _sim_results_files):
        overwrite_alert_open = True
        load_button_disabled = False
    elif len(name) <= 0:
        overwrite_alert_open = False
        load_button_disabled = True

    return overwrite_alert_open, load_button_disabled, name


## PLOT CONFIGURATION --------------------------------------------------------
@app.callback(
    Output('plot-config-upload', 'contents'),
    Output('plot-config-general-store', 'data'),
    Output('plot-config-x-store', 'data'),
    Output('plot-config-y-store', 'data'),
    Output({'component': 'error-box', 'id': 'update-config'}, 'is_open'),
    Output({'component': 'error-box-text', 'id': 'update-config-text'}, 'children'),
    Input('plot-config-upload', 'contents'),
    Input('data-file-store', 'data'),
    Input({'component': 'plot-config', 'tab': 'general', 'field': 'title'}, 'value'),                # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'general', 'field': 'background-color'}, 'value'),     # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'general', 'field': 'grid-x'}, 'value'),               # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'general', 'field': 'grid-y'}, 'value'),               # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'general', 'field': 'append-units'}, 'value'),         # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'general', 'field': 'width-per-y-axis'}, 'value'),     # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'general', 'field': 'freeze-uirevision'}, 'value'),    # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'general', 'field': 'hovermode'}, 'value'),            # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'general', 'field': 'image_export_type'}, 'value'),    # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'general', 'field': 'image_export_width'}, 'value'),   # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'general', 'field': 'image_export_height'}, 'value'),  # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'general', 'field': 'image_export_scale'}, 'value'),   # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'x', 'field': 'variable'}, 'value'),
    Input({'component': 'plot-config', 'tab': 'x', 'field': 'units'}, 'value'),
    Input({'component': 'plot-config', 'tab': 'x', 'field': 'title'}, 'value'),
    Input({'component': 'plot-config', 'tab': 'x', 'field': 'xmin'}, 'value'),
    Input({'component': 'plot-config', 'tab': 'x', 'field': 'xmax'}, 'value'),
    Input({'component': 'plot-config', 'tab': 'x', 'field': 'tick_spacing'}, 'value'),  # noqa: E501  # pylint: disable=C0301
    Input('add-y-axis-button', 'n_clicks'),
    Input('delete-y-axis-button', 'n_clicks'),
    Input('hide-show-y-axis-button', 'n_clicks'),
    Input('add-y-data-button', 'n_clicks'),
    Input('delete-y-data-button', 'n_clicks'),
    Input('hide-show-y-data-button', 'n_clicks'),
    Input({'component': 'plot-config', 'tab': 'y', 'field': 'title'}, 'value'),
    Input({'component': 'plot-config', 'tab': 'y', 'field': 'axis-color'}, 'value'),  # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'y', 'field': 'ymin'}, 'value'),
    Input({'component': 'plot-config', 'tab': 'y', 'field': 'ymax'}, 'value'),
    Input({'component': 'plot-config', 'tab': 'y', 'field': 'tick_spacing'}, 'value'),    # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'y', 'field': 'legend-title'}, 'value'),    # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'y', 'field': 'trace-file'}, 'value'),      # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'y', 'field': 'trace-variable'}, 'value'),  # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'y', 'field': 'units'}, 'value'),           # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'y', 'field': 'trace-color'}, 'value'),     # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'y', 'field': 'trace-width'}, 'value'),     # noqa: E501  # pylint: disable=C0301
    Input({'component': 'plot-config', 'tab': 'y', 'field': 'trace-dash'}, 'value'),      # noqa: E501  # pylint: disable=C0301
    State('plot-config-general-store', 'data'),
    State('plot-config-x-store', 'data'),
    State('plot-config-y-store', 'data'),
    State({'component': 'error-box', 'id': dash.ALL}, 'is_open'),
    State('y-axis-selector', 'active_page'),
    State('y-data-selector', 'active_page'),
    prevent_initial_call=True,
)
def update_plot_config(
    ## Inputs ##
    upload_contents: Optional[str],
    file_metadata: dict,
    plot_title: Optional[str],
    background_color: Optional[str],
    grid_x_enabled: Optional[bool],
    grid_y_enabled: Optional[bool],
    append_units: Optional[bool],
    width_per_y_axis: Optional[str],
    freeze_uirevision: Optional[bool],
    hovermode: Optional[str],
    image_export_type,
    image_export_width,
    image_export_height,
    image_export_scale,
    x_variable: Optional[str],
    x_units: Optional[str],
    x_title: Optional[str],
    x_min,
    x_max,
    x_tick_spacing,
    add_y_axis_clicks: int,
    delete_y_axis_clicks: int,
    hide_show_y_axis_clicks: int,
    add_trace_clicks: int,
    delete_trace_clicks: int,
    hide_show_trace_clicks: int,
    y_title: Optional[str],
    y_color: Optional[str],
    y_min,
    y_max,
    y_tick_spacing,
    trace_name: Optional[str],
    trace_file: Optional[str],
    trace_variable: Optional[str],
    trace_units: Optional[str],
    trace_color: str,
    trace_width,
    trace_dash: str,
    ## States ##
    config_general: dict,
    config_x: dict,
    config_y: dict,
    error_boxes_open: List[bool],
    selected_y_axis_page: int,
    selected_data_series_page: int,
):
    """Updates the browser storage containing the plot configuration data"""
    if any(error_boxes_open):
        # If any error boxes are open, prevent updating the app so user can
        # resolve the issue first
        raise dash.exceptions.PreventUpdate

    # Determine which axis and data series are selected displayed to the user
    # (these are the objects whose settings may be edited)
    y_axis_idx = selected_y_axis_page - 1
    trace_idx = selected_data_series_page - 1

    if dash.ctx.triggered_id == 'plot-config-upload':
        if upload_contents is None:
            # Either callback was run in initial call, or it was already
            # triggered and automatically reset the upload "contents" to None.
            # In either case, no update is necessary
            raise dash.exceptions.PreventUpdate

        try:
            new_general, new_x, new_y = load_plot_config(upload_contents)
            update_graph(config_general=new_general, config_x=new_x,
                         config_y=new_y, sim_results_files=_sim_results_files,
                         file_metadata=file_metadata)
        except Exception as exception:
            error_text = generate_error_box_text(load_plot_config_error_message,
                                                 exception)

            return None, config_general, config_x, config_y, True, error_text

        config_general = new_general
        config_x = new_x
        config_y = new_y

    # If simulation results files were uploaded or deleted, make appropriate
    # adjustments to configuration options
    elif dash.ctx.triggered_id == 'data-file-store':
        # If a new file was uploaded and it doesn't contain the x-axis
        # variable, reset the x-axis variable
        if not all(config_x['variable'] in file.variables
                   for file in _sim_results_files.values()):
            config_x['variable'] = None
            config_x['units'] = None

        # If a file was deleted, remove any associated traces
        for y_axis in config_y['axes']:
            for i, trace in enumerate(y_axis['traces']):
                if trace['file'] not in _sim_results_files.keys():
                    del y_axis['traces'][i]

    # Note that for the callback actions below, the buttons' "n_clicks"
    # properties are reset to 0 when the configuration panel is rendered, so
    # checking that "n_clicks > 0" is necessary (or else the branch will be
    # run when the panel is rendered and "n_clicks" is reset from 1 to 0).

    # Detect and respond to button presses to manage y-axes
    elif dash.ctx.triggered_id == 'add-y-axis-button':
        if add_y_axis_clicks > 0:
            config_y['axes'].append(copy.deepcopy(default_y_axis_settings))
        else:
            # Callback is fired when components are dynamically generated, so
            # if this is the case, skip running actions
            raise dash.exceptions.PreventUpdate

    elif dash.ctx.triggered_id == 'delete-y-axis-button':
        if delete_y_axis_clicks > 0:
            del config_y['axes'][y_axis_idx]
        else:
            raise dash.exceptions.PreventUpdate

    elif dash.ctx.triggered_id == 'hide-show-y-axis-button':
        if hide_show_y_axis_clicks > 0:
            config_y['axes'][y_axis_idx]['enabled'] \
                = not config_y['axes'][y_axis_idx]['enabled']
        else:
            raise dash.exceptions.PreventUpdate

    # Detect and respond to button presses to manage data series
    elif dash.ctx.triggered_id == 'add-y-data-button':
        if add_trace_clicks > 0:
            config_y['axes'][y_axis_idx]['traces'].append(
                copy.deepcopy(default_trace_settings))
            config_y['axes'][y_axis_idx]['traces'][-1]['style']['color'] \
                = random_hex_color()
        else:
            raise dash.exceptions.PreventUpdate

    elif dash.ctx.triggered_id == 'delete-y-data-button':
        if delete_trace_clicks > 0:
            del config_y['axes'][y_axis_idx]['traces'][trace_idx]
        else:
            raise dash.exceptions.PreventUpdate

    elif dash.ctx.triggered_id == 'hide-show-y-data-button':
        if hide_show_trace_clicks > 0:
            config_y['axes'][y_axis_idx]['traces'][trace_idx]['enabled'] \
                = not config_y['axes'][y_axis_idx]['traces'][trace_idx]['enabled']
        else:
            raise dash.exceptions.PreventUpdate

    # Update stored properties if any of the dropdowns or input boxes were changed
    else:
        # General plot configuration settings
        config_general['title'] = plot_title

        if background_color is not None:
            config_general['background_color'] = background_color

        if grid_x_enabled is not None:
            config_general['grid_x'] = grid_x_enabled

        if grid_y_enabled is not None:
            config_general['grid_y'] = grid_y_enabled

        if append_units is not None:
            config_general['append_units'] = append_units

        if width_per_y_axis is not None:
            config_general['width_per_y_axis'] = float(width_per_y_axis)

        if hovermode is not None:
            config_general['hovermode'] = hovermode

        if freeze_uirevision is not None:
            config_general['freeze_uirevision'] = freeze_uirevision

        if image_export_type is not None:
            config_general['image_export_type'] = image_export_type

        if image_export_width not in (None, ''):
            config_general['image_export_width'] = int(float(image_export_width))
        else:
            config_general['image_export_width'] = None

        if image_export_height not in (None, ''):
            config_general['image_export_height'] = int(float(image_export_height))
        else:
            config_general['image_export_height'] = None

        if image_export_scale in (None, ''):
            image_export_scale = 1
        config_general['image_export_scale'] = int(float(image_export_scale))

        # Plot configuration settings for x-axis
        if x_variable != config_x['variable']:
            if x_variable in ('', None):
                x_units = None
            else:
                x_units = (_sim_results_files[list(_sim_results_files.keys())[0]]
                           .get_units(x_variable))

        config_x['variable'] = x_variable
        config_x['units'] = x_units
        config_x['axis_title'] = x_title
        config_x['xmin'] = None if x_min in (None, '') else float(x_min)
        config_x['xmax'] = None if x_max in (None, '') else float(x_max)
        config_x['tick_spacing'] = None if x_tick_spacing in (None, '') \
                                        else float(x_tick_spacing)  # noqa: E127

        # Plot configuration settings for y-axes
        if len(config_y['axes']) > 0:
            y_axis = config_y['axes'][y_axis_idx]

            y_axis['axis_title'] = y_title

            if y_color is not None:
                y_axis['color'] = y_color

            y_axis['ymin'] = None if y_min in (None, '') else float(y_min)
            y_axis['ymax'] = None if y_max in (None, '') else float(y_max)
            y_axis['tick_spacing'] = None if y_tick_spacing in (None, '') \
                                          else float(y_tick_spacing)  # noqa: E127

            if len(config_y['axes'][y_axis_idx]['traces']) > 0:
                trace = config_y['axes'][y_axis_idx]['traces'][trace_idx]

                if trace_variable != trace['variable']:
                    if trace_variable in ('', None):
                        trace_units = None
                    else:
                        trace_units = (_sim_results_files[trace_file]
                                       .get_units(trace_variable))

                trace['name'] = trace_name
                trace['file'] = trace_file
                trace['variable'] = trace_variable
                trace['units'] = trace_units
                trace['style']['color'] = trace_color
                trace['style']['width'] = None if trace_width in (None, '') \
                                               else float(trace_width)  # noqa: E127
                trace['style']['dash'] = trace_dash

    return None, config_general, config_x, config_y, False, None


@app.callback(
    Output('config-export-download', 'data'),
    Input('config-export-button', 'n_clicks'),
    State('plot-config-general-store', 'data'),
    State('plot-config-x-store', 'data'),
    State('plot-config-y-store', 'data'),
    prevent_initial_call=True,
)
def export_plot_config(n_clicks: Optional[int], config_general: dict,
                       config_x: dict, config_y: dict):
    """Saves the current plot configuration options in a file and makes it
    available to the user as a download"""
    data = {
        'base64': False,
        'content': plot_config_to_str(config_general, config_x, config_y),
        'filename': 'mahautils_simviewer_config.json',
    }
    return data


@app.callback(
    Output('plot-config-general-settings', 'children'),
    Input('plot-config-general-store', 'data'),
)
def render_ui_general(config_general: dict):
    """Generates the UI elements that allow the user to configure general
    plot settings"""
    return render_general_settings(config_general)


@app.callback(
    Output('plot-config-x-settings', 'children'),
    Input('plot-config-x-store', 'data'),
)
def render_ui_x(config_x: dict):
    """Generates the UI elements that allow the user to configure x-axis
    plot settings"""
    return render_x_settings(config_x, _sim_results_files)


@app.callback(
    Output('plot-config-y-settings', 'children'),
    Output('y-data-selector', 'active_page'),
    Input('plot-config-y-store', 'data'),
    Input('data-file-store', 'data'),
    Input('y-axis-selector', 'active_page'),
    Input('y-data-selector', 'active_page'),
    Input('add-y-axis-button', 'n_clicks'),
    Input('add-y-data-button', 'n_clicks'),
)
def render_ui_y(config_y: dict, file_metadata: dict,
                selected_axis_page: int, selected_trace_page: int,
                add_y_axis_clicks: int, add_y_data_clicks: int):
    """Generates the UI elements that allow the user to configure y-axis
    plot settings"""
    if dash.ctx.triggered_id == 'y-axis-selector':
        # If the user switched y-axes, reset selected trace to first trace
        # (otherwise, the currently selected trace might be out of range)
        selected_trace_page = 1

    # If the user just created a new y-axis or data series, display it
    num_axes = len(config_y['axes'])
    if dash.ctx.triggered_id == 'add-y-axis-button':
        selected_axis_page = num_axes

    if dash.ctx.triggered_id == 'add-y-data-button':
        num_traces = len(config_y['axes'][selected_axis_page-1]['traces'])
        selected_trace_page = num_traces

    # Ensure that the selected axis and trace are within a valid range (they
    # might not be if an axis or trace was just deleted)
    selected_axis_page = max(1, min(selected_axis_page, num_axes))

    selected_trace_page = max(1, selected_trace_page)
    if num_axes > 0:
        num_traces = len(config_y['axes'][selected_axis_page-1]['traces'])
        selected_trace_page = min(selected_trace_page, num_traces)

    layout = render_y_settings(
        config_y, _sim_results_files, file_metadata,
        selected_axis_page-1 if selected_axis_page is not None else 0,
        selected_trace_page-1 if selected_trace_page is not None else 0,
    )

    return layout, selected_trace_page


## PLOTTING ------------------------------------------------------------------
@app.callback(
    Output('plotly-graph', 'figure'),
    Output('plotly-graph', 'config'),
    Output({'component': 'error-box', 'id': 'update-plot'}, 'is_open'),
    Output({'component': 'error-box-text', 'id': 'update-plot-text'}, 'children'),
    Input('plot-config-general-store', 'data'),
    Input('plot-config-x-store', 'data'),
    Input('plot-config-y-store', 'data'),
    Input('data-file-store', 'data'),
    State('plotly-graph', 'figure'),
    State({'component': 'error-box', 'id': dash.ALL}, 'is_open'),
    State('plotly-graph', 'config'),
    prevent_initial_call=True,
)
def update_plot(
    ## Inputs ##
    config_general: dict,
    config_x: dict,
    config_y: dict,
    file_metadata: dict,
    ## States ##
    current_figure: go.Figure,
    error_boxes_open: List[bool],
    figure_config: dict,
):
    """Updates the simulation results plot any time the plot configuration
    data dictionary changes"""
    if any(error_boxes_open):
        # If any error boxes are open, prevent updating the app so user can
        # resolve the issue first
        raise dash.exceptions.PreventUpdate

    try:
        figure = update_graph(
            config_general    = config_general,
            config_x          = config_x,
            config_y          = config_y,
            sim_results_files = _sim_results_files,
            file_metadata     = file_metadata,
        )

        figure_config['toImageButtonOptions']['format'] \
            = config_general['image_export_type']

        figure_config['toImageButtonOptions']['width'] \
            = config_general['image_export_width']

        figure_config['toImageButtonOptions']['height'] \
            = config_general['image_export_height']

        figure_config['toImageButtonOptions']['scale'] \
            = config_general['image_export_scale']

        error_box_is_open = False
        error_box_text = None

    except Exception as exception:
        figure = current_figure
        error_box_is_open = True
        error_box_text = generate_error_box_text(
            dash.html.P('Unable to update plot.  There is likely a mistake '
                        'in the plot configuration options you have selected.'),
            exception,
        )

    return figure, figure_config, error_box_is_open, error_box_text


## HIDE/SHOW CONTROLS FOR CONFIGURATION PANES --------------------------------
@app.callback(
    Output('info-button-modal', 'is_open'),
    Input('info-button', 'n_clicks'),
    Input('info-box-close-button', 'n_clicks'),
    State('info-button-modal', 'is_open'),
    prevent_initial_call=True,
)
def toggle_info_box(n_clicks_open, n_clicks_close, is_open):
    """Callback which opens and closes the modal information box
    """
    return not is_open


@app.callback(
    Output('plot-config-panel', 'is_open'),
    Input('plot-config-button', 'n_clicks'),
    State('plot-config-panel', 'is_open'),
    prevent_initial_call=True,
)
def toggle_plot_config_panel(n_clicks, is_open):
    """Callback which opens and closes the plot configuration panel
    """
    return not is_open



"""Configuration panel tab for uploading and managing simulation results
data files.
"""

# Mypy type checking disabled for packages that are not PEP 561-compliant
import dash                              # type: ignore
import dash_bootstrap_components as dbc  # type: ignore

from mahautils.multics.sim_results_viewer.constants import (
    TAB_BAR_PADDING,
)
from .utils_generate_table import empty_file_table


def data_files_tab():
    """Creates the configuration panel tab where simulation results files can
    be uploaded and managed"""
    return [
        dash.html.H5('Data Upload', style={'marginTop': TAB_BAR_PADDING}),
        dash.html.P('Upload a new simulation results file here.'),
        dbc.Spinner(
            dash.dcc.Upload(
                'Drag and Drop or Click to Browse',
                id='upload-data',
                style={
                    'width': '100%',
                    'height': '60px',
                    'lineHeight': '60px',
                    'borderWidth': '1px',
                    'borderStyle': 'dashed',
                    'borderRadius': '10px',
                    'textAlign': 'center',
                    'marginLeft': '0px',
                    'marginRight': '0px',
                },
            ),
            # Delay prevents showing spinner when using switches or delete button
            delay_show=200,
        ),
        _file_name_input_box(),
        dash.html.Hr(),
        dash.html.H5('Current Files'),
        dash.html.P('Manage your uploaded simulation results files here.'),
        empty_file_table(),
    ]


def _file_name_input_box():
    return dbc.Spinner(
        dash.html.Div(
            dbc.Stack([
                # File name input box
                dbc.Row([
                    dbc.Col(
                        dbc.Input(
                            placeholder='Enter a name for your file...',
                            id='user-file-name',
                        ),
                        width=9, style={'marginTop': '10px'},
                    ),
                    dbc.Col(
                        dbc.Button('Load', id='load-file-button', disabled=False),
                        width=3, style={'marginTop': '10px'},
                    ),
                ]),

                # Overwrite warning alert
                dbc.Row(
                    dbc.Alert(
                        [
                            dash.html.H5([
                                dash.html.I(className='fa fa-exclamation-triangle'),
                                ' WARNING',
                            ]),
                            dash.html.Hr(),
                            dash.html.P(
                                'A file with this name already exists and '
                                'will be overwritten if you click "Load"'
                            ),
                        ],
                        id='file-overwrite-alert',
                        color='warning',
                        style={
                            'width': '95%',
                            'marginTop': '10px',
                            'marginLeft': '10px',
                            'marginRight': '10px',
                        },
                        is_open=False,
                        dismissable=True,
                    ),
                ),
            ]),
            id='div-user-file-name',
            hidden=True,
        ),
        # Delay prevents showing spinner when clearing "contents" (after
        # loading simulation results file)
        delay_show=500,
    )



"""Constants used in the Maha Multics software and related input/output files
are defined in this module, so they can be used for unit conversions and other
computations in the :py:mod:`mahautils.multics` module.
"""

# Atmospheric pressure
ATMOSPHERIC_PRESSURE_PA = 101325    # Pa (absolute pressure)


# Standard acceleration due to gravity
# https://physics.nist.gov/cgi-bin/cuu/Value?gn
STANDARD_GRAVITY = 9.80665  # m/s^2


# Conversion factors
# Inches to meters
_IN_TO_M_SCALE = 0.0254  # m/in

# Pound-mass to kilograms
_LBM_TO_KG_SCALE = 0.45359237  # kg/lbm

# Pound-force to Newtons
_LBF_TO_N_SCALE = 4.4482216152605  # N/lbf

# Pounds per square inch (psi) to pascals (Pa)
_PSI_TO_PA_SCALE = (_LBM_TO_KG_SCALE * STANDARD_GRAVITY) \
    / (_IN_TO_M_SCALE**2)  # Pa/psi

# Degrees Celsius to Kelvin
_DEG_C_TO_K_OFFSET = 273.15

# Degrees Fahrenheit to Kelvin
_DEG_F_TO_K_SCALE = 5/9
_DEG_F_TO_K_OFFSET = -32 * _DEG_F_TO_K_SCALE + _DEG_C_TO_K_OFFSET



"""Panel tab for configuring x-axis plot settings.
"""

# Mypy type checking disabled for packages that are not PEP 561-compliant
import dash                              # type: ignore
import dash_bootstrap_components as dbc  # type: ignore

from mahautils.multics.sim_results_viewer.constants import (
    SIM_RESULTS_DICT_T,
    SIM_VAR_DROPDOWN_OPTION_HEIGHT,
    TAB_BAR_PADDING,
    UI_DESCRIPTION_MARGIN_BELOW,
)
from mahautils.multics.sim_results_viewer.store import default_plot_config_x
from mahautils.multics.sim_results_viewer.ui_elements import tick_range_step


def plot_settings_x():
    """Creates the configuration panel tab with settings for the x-axis"""
    return [
        dash.html.Div(
            render_x_settings(default_plot_config_x, {}),
            id='plot-config-x-settings',
            style={'marginTop': TAB_BAR_PADDING},
        ),
    ]


def render_x_settings(config_x: dict, sim_results_files: SIM_RESULTS_DICT_T):
    """Generates the UI elements that allow the user to configure x-axis
    plot settings"""
    # Identify list of simulation results variables that are present and have
    # the same description in all loaded simulation results files
    variable_options = {}

    for i, (file_name, sim_results) in enumerate(sim_results_files.items()):
        if i == 0:
            key0 = file_name
            variable_options = {
                var: {
                    'label': f'{var} — {sim_results.get_description(var)}',
                    'value': var,
                }
                for var in sim_results.variables
            }
        else:
            for var in list(variable_options.keys()):
                # If a given variable is not present in all simulation results
                # files or it has different descriptions, remove it from the
                # list of valid options
                if var not in sim_results.variables:
                    del variable_options[var]
                elif (
                    sim_results_files[key0].get_description(var)
                    != sim_results.get_description(var)
                ):
                    del variable_options[var]

    return [
        dash.html.H4('Data'),
        dash.html.Hr(),
        dash.html.H5('Simulation Results Variable'),
        dash.html.P(
            ('Sets the simulation results variable to be plotted '
             'on the horizontal axis for all simulation results'),
            style={'marginBottom': UI_DESCRIPTION_MARGIN_BELOW},
        ),
        dash.html.Div(dbc.Row([
            dbc.Col(
                dash.html.H6('Variable'),
                width=2,
            ),
            dbc.Col(
                dash.dcc.Dropdown(
                    options=list(variable_options.values()),
                    value=config_x['variable'],
                    id={'component': 'plot-config', 'tab': 'x',
                        'field': 'variable'},
                    clearable=True,
                    multi=False,
                    optionHeight=SIM_VAR_DROPDOWN_OPTION_HEIGHT,
                ),
                width=10,
            ),
        ]), style={'marginBottom': UI_DESCRIPTION_MARGIN_BELOW}),
        dbc.Row([
            dbc.Col(
                dash.html.H6('Units'),
                width=2,
            ),
            dbc.Col(
                dbc.Input(
                    debounce=True,
                    value=config_x['units'],
                    id={'component': 'plot-config', 'tab': 'x',
                        'field': 'units'},
                    style={'width': 150},
                    disabled=(config_x['variable'] is None),
                ),
                width=10,
            ),
        ]),
        dbc.Alert(
            [
                dash.html.H5([
                    dash.html.I(className='fa fa-exclamation-triangle'),
                    ' WARNING',
                ]),
                dash.html.Hr(),
                dash.html.P('The plot will not be updated until you have '
                            'selected both a simulation results variable AND '
                            'units for the x-axis'),
            ],
            color='danger',
            style={
                'width': '95%',
                'marginTop': '10px',
                'marginLeft': '10px',
                'marginRight': '10px',
            },
            is_open=(
                config_x['variable'] in (None, '')
                or config_x['units'] in (None, '')
            ),
            dismissable=False,
        ),
        dash.html.Br(),
        dash.html.Br(),
        dash.html.H4('Formatting'),
        dash.html.Hr(),
        dash.html.H5('Axis Title'),
        dbc.Input(
            debounce=True,
            id={'component': 'plot-config', 'tab': 'x',
                'field': 'title'},
            placeholder='Enter a title for the x-axis...',
            value=config_x['axis_title'],
        ),
        dash.html.Br(),
        tick_range_step(
            range_min_id={'component': 'plot-config', 'tab': 'x',
                          'field': 'xmin'},
            range_max_id={'component': 'plot-config', 'tab': 'x',
                          'field': 'xmax'},
            tick_spacing_id={'component': 'plot-config', 'tab': 'x',
                             'field': 'tick_spacing'},
            range_min_val=config_x['xmin'],
            range_max_val=config_x['xmax'],
            tick_spacing_val=config_x['tick_spacing'],
        ),
    ]



"""Classes for representing a polygon

The classes in this module are intended to represent a polygon in the 2D
Cartesian coordinate system.
"""

import copy
import math
from typing import List, Optional, Tuple, Union

# Mypy type checking disabled for packages that are not PEP 561-compliant
import matplotlib.path  # type: ignore
import numpy as np

from .point import Array_Float2, Point
from .point2D import CartesianPoint2D
from .shape_open_closed import ClosedShape2D

ListOfPoints2D = Union[
    List[Array_Float2], Tuple[Array_Float2, ...], np.ndarray,
    List[Point], Tuple[Point, ...],
]


class Polygon(ClosedShape2D):
    """An object representing a closed polygon in the 2D Cartesian plane

    This class is intended to represent a polygon, comprised of the closed
    region inside a series of vertices connected by straight lines.  The
    locations of the vertices are arbitrary and edges of the polygon can
    self-intersect.

    Examples
    --------
    Create a triangle:

    >>> vertices = [[0, 0], [1, 0], [1, 2]]
    >>> triangle = mahautils.shapes.Polygon(vertices)

    View the points on the boundary of the polygon, optionally repeating the
    start/end point of the polygon:

    >>> triangle.points()
    (array([0., 0.]), array([1., 0.]), array([1., 2.]))
    >>> triangle.points(repeat_end=True)
    (array([0., 0.]), array([1., 0.]), array([1., 2.]), array([0., 0.]))

    Check whether a point is inside the polygon:

    >>> triangle.is_inside((0.5, 0.5))
    True
    >>> triangle.is_inside((0.5, 10))
    False
    """

    def __init__(self, vertices: ListOfPoints2D,
                 construction: bool = False,
                 polygon_file_enclosed_conv: int = 1,
                 units: Optional[str] = None) -> None:
        """Creates an object representing a polygon

        Parameters
        ----------
        vertices : list or tuple or np.ndarray
            A 2D array containing a set of 2D points representing the vertices
            of the polygon, in order
        construction : bool, optional
            Whether the shape is a "construction shape" meant for visual
            display but not functional geometry (default is ``False``)
        polygon_file_enclosed_conv : int, optional
            Convention for considering enclosed area when generating a polygon
            file from :py:class:`ClosedShape2D` objects (default is ``1``)
        units : str, optional
            The units in which the geometry is defined, or ``None`` to
            indicate dimensionless geometry or that units are to be ignored
            (default is ``None``)
        """
        # Mypy disabled as a workaround for python/mypy#3004
        self.vertices = vertices  # type: ignore

        super().__init__(
            default_num_coordinates=None,
            construction=construction,
            polygon_file_enclosed_conv=polygon_file_enclosed_conv,
            units=units,
        )

    @property
    def vertices(self) -> np.ndarray:
        """Returns a copy of a NumPy array containing the vertices of the
        polygon

        Even though polygons are a closed shape, points are stored in the
        :py:attr:`vertices` attribute such that the first and last point are
        **not** repeated.
        """
        return copy.deepcopy(self._vertices)

    @vertices.setter
    def vertices(self, vertices: ListOfPoints2D) -> None:
        try:
            # Convert "points" argument to NumPy array
            vertices_array = np.array([list(x) for x in vertices],
                                      dtype=np.float64)

            # Verify expected shape
            if (vertices_array.ndim != 2) or (vertices_array.shape[1] != 2):
                raise ValueError('Polygon vertices are not a 2D array')

        except (TypeError, ValueError) as exception:
            raise ValueError(
                'Invalid polygon vertex format. Polygon vertices must be a '
                'list of 2D points') from exception

        # If user repeated the first/last vertex, remove it before storing
        # internally
        if np.array_equal(vertices_array[0], vertices_array[-1]):
            vertices_array = vertices_array[:-1]

        self._vertices = vertices_array

        # Store Matplotlib Path object representing polygon
        self._matplotlib_path = matplotlib.path.Path(self._vertices)

    def is_inside(self, point: Union[Array_Float2, CartesianPoint2D],
                  perimeter_is_inside: bool = True) -> bool:
        if perimeter_is_inside:
            # Python stores floating-point numbers as doubles, so the decimal
            # precision is ~15 decimal places
            tolerance = 1e-15

            # Check whether point is within limits of floating-point precision
            # from border
            return any([
                self._matplotlib_path.contains_point(
                    point=list(point), radius=tolerance),
                self._matplotlib_path.contains_point(
                    point=list(point), radius=-tolerance),
            ])

        return self._matplotlib_path.contains_point(point=list(point), radius=0)

    def points(self, repeat_end: bool = False) -> Tuple[np.ndarray, ...]:
        return self._convert_xy_coordinates_to_points(repeat_end=repeat_end)

    def rotate(self, center: Union[Array_Float2, CartesianPoint2D],
               angle: float, angle_units: str = 'rad') -> None:
        # Convert angle to radians
        angle = self._convert_rotate_angle(angle, angle_units)

        # Shift geometry to origin
        center = CartesianPoint2D(center)
        self.translate(x=(-center.x), y=(-center.y))

        # Rotate geometry about origin
        rotation_matrix = np.array([
            [ math.cos(angle), math.sin(angle)],  # noqa: E201
            [-math.sin(angle), math.cos(angle)],
        ])
        self._vertices = np.matmul(self._vertices, rotation_matrix)

        # Shift geometry back to center of rotation
        self.translate(x=center.x, y=center.y)

    def translate(self, x: float = 0, y: float = 0) -> None:
        self._vertices[:, 0] += x
        self._vertices[:, 1] += y

    def xy_coordinates(self, repeat_end: bool = False
                       ) -> Tuple[np.ndarray, np.ndarray]:
        points = self.vertices

        if repeat_end:
            points = np.concatenate([points, [points[0]]], axis=0)

        return (points[:, 0], points[:, 1])



"""This module provides a class intended for reading and parsing fluid
property files used by the Maha Multics software.
"""

import copy
import pathlib
from typing import List, Optional, Tuple, Union

# Mypy type checking disabled for packages that are not PEP 561-compliant
import numpy as np
import pyxx
import scipy.interpolate  # type: ignore

from mahautils.utils.arrays import to_np_1D_array
from .configfile import MahaMulticsConfigFile
from .exceptions import (
    FileNotParsedError,
    FluidPropertyFileError,
)

NumberOrNumericList = Union[float, np.ndarray, List[float], Tuple[float, ...]]


class FluidPropertyFile(MahaMulticsConfigFile):
    """An object representing a Maha Multics fluid properties file

    The Maha Multics software requires fluid properties to simulate
    lubrication behavior.  One method of providing fluid properties is with
    formatted fluid property files.  This class allows such files to be read
    so their properties can be viewed.

    Maha Multics fluid property files store the following properties:

    - Density
    - Bulk modulus
    - Kinematic viscosity
    - Specific heat capacity
    - Thermal conductivity
    - Volumetric expansion coefficient
    - Specific enthalpy
    """

    # Order and units of columns of data in fluid property files
    __file_format = (
        # Symbol   Description                         Units     Attribute name
        ('rho',    'density',                          'kg/m^3', '_density'      ),  # noqa: E202, E203, E501  # pylint: disable=C0301
        ('k',      'bulk modulus',                     'Pa_a'  , '_bulk_modulus' ),  # noqa: E202, E203, E501  # pylint: disable=C0301
        ('nu',     'kinematic viscosity',              'm^2/s' , '_viscosity_k'  ),  # noqa: E202, E203, E501  # pylint: disable=C0301
        ('cp',     'specific heat capacity',           'J/kg/K', '_specific_heat'),  # noqa: E202, E203, E501  # pylint: disable=C0301
        ('lambda', 'thermal conductivity',             'W/m/K' , '_thermal_cond' ),  # noqa: E202, E203, E501  # pylint: disable=C0301
        ('alpha',  'volumetric expansion coefficient', 'K^(-1)', '_expand_coeff' ),  # noqa: E202, E203, E501  # pylint: disable=C0301
        ('h',      'specific enthalpy',                'J/kg'  , '_enthalpy'     ),  # noqa: E202, E203, E501  # pylint: disable=C0301
    )

    # Default units of pressure and temperature in fluid property files
    __pressure_units = 'Pa_a'  # Pascals, absolute pressure
    __temperature_units = 'K'  # Kelvin

    def __init__(self, path: Optional[Union[str, pathlib.Path]] = None,
                 unit_converter: Optional[pyxx.units.UnitConverter] = None
                 ) -> None:
        """Creates an object that can read and write Maha Multics fluid
        property files

        Creates an instance of the :py:class:`FluidPropertyFile` class and
        optionally reads and parses a specified Maha Multics fluid property
        file.

        Parameters
        ----------
        path : str or pathlib.Path, optional
            The path and filename of the fluid property file to read and
            parse (default is ``None``).  If set to ``None``, no file is read
        unit_converter : pyxx.units.UnitConverter, optional
            A :py:class:`pyxx.units.UnitConverter` instance which will be
            used to convert units of quantities stored in the simulation
            results file (default is ``None``).  If set to ``None``, the
            :py:class:`MahaMulticsUnitConverter` unit converter will be used
            to perform unit conversions
        """
        super().__init__(path=path, unit_converter=unit_converter)

        # Initialize variables
        self._num_pressure: Union[int, None] = None
        self._num_temperature: Union[int, None] = None
        self._step_pressure: Union[float, None] = None
        self._step_temperature: Union[float, None] = None
        self._pressure_values: Union[np.ndarray, None] = None
        self._temperature_values: Union[np.ndarray, None] = None

        self._density: Union[np.ndarray, None] = None
        self._bulk_modulus: Union[np.ndarray, None] = None
        self._viscosity_k: Union[np.ndarray, None] = None
        self._specific_heat: Union[np.ndarray, None] = None
        self._thermal_conduct: Union[np.ndarray, None] = None
        self._expand_coeff: Union[np.ndarray, None] = None
        self._enthalpy: Union[np.ndarray, None] = None

        # If path was provided, read file
        if path is not None:
            self.read(path, parse=True)

    @property
    def num_pressure(self) -> int:
        """The number of (discrete) values of pressure for which fluid
        properties are defined in the file"""
        if self._num_pressure is None:
            raise FileNotParsedError('Attribute "num_pressure" is not defined; '
                                     'file has not yet been parsed')

        return self._num_pressure

    @property
    def num_temperature(self) -> int:
        """The number of (discrete) values of temperature for which fluid
        properties are defined in the file"""
        if self._num_temperature is None:
            raise FileNotParsedError('Attribute "num_temperature" is not '
                                     'defined; file has not yet been parsed')

        return self._num_temperature

    def get_pressure_step(self, units: str) -> float:
        """Returns the increments of pressure over which properties in the
        file are stored

        Fluid properties are stored in discrete increments of pressure.  This
        method returns this "step" or increment between pressure values.

        Parameters
        ----------
        units : str
            The units in which the pressure increment should be stored

        Returns
        -------
        float
            The pressure increment over which fluid properties are stored
        """
        if self._step_pressure is None:
            raise FileNotParsedError(
                'Cannot return pressure step; file has not yet been parsed')

        return float(self.unit_converter.convert(
            quantity  = self._step_pressure,
            from_unit = self.__pressure_units,
            to_unit   = units,
        ))

    def get_pressure_values(self, units: str) -> np.ndarray:
        """Returns the (discrete) pressure values over which properties in the
        file are stored

        Fluid properties are stored for discrete values of pressure.  This
        method returns all pressure values for which fluid properties are
        stored.

        Parameters
        ----------
        units : str
            The units in which the pressure values should be stored

        Returns
        -------
        np.ndarray
            The pressure values for which fluid properties are stored
        """
        if self._pressure_values is None:
            raise FileNotParsedError(
                'Cannot return pressure values; file has not yet been parsed')

        return self.unit_converter.convert(
            quantity  = self._pressure_values,
            from_unit = self.__pressure_units,
            to_unit   = units,
        )

    def get_temperature_step(self, units: str) -> float:
        """Returns the increments of temperature over which properties in the
        file are stored

        Fluid properties are stored in discrete increments of temperature.
        This method returns this "step" or increment between temperature
        values.

        Parameters
        ----------
        units : str
            The units in which the temperature increment should be returned

        Returns
        -------
        float
            The temperature increment over which fluid properties are stored
        """
        if self._step_temperature is None:
            raise FileNotParsedError(
                'Cannot return temperature step; file has not yet been parsed')

        return float(self.unit_converter.convert(
            quantity  = self._step_temperature,
            from_unit = self.__temperature_units,
            to_unit   = units,
        ))

    def get_temperature_values(self, units: str) -> np.ndarray:
        """Returns the (discrete) temperature values over which properties in
        the file are stored

        Fluid properties are stored for discrete values of temperature.  This
        method returns all temperature values for which fluid properties are
        stored.

        Parameters
        ----------
        units : str
            The units in which the temperature values should be stored

        Returns
        -------
        np.ndarray
            The temperature values for which fluid properties are stored
        """
        if self._temperature_values is None:
            raise FileNotParsedError(
                'Cannot return temperature values; file has not yet been parsed')

        return self.unit_converter.convert(
            quantity  = self._temperature_values,
            from_unit = self.__temperature_units,
            to_unit   = units,
        )

    def interpolate(self, fluid_property: str, output_units: str,
                    pressures: NumberOrNumericList, pressure_units: str,
                    temperatures: NumberOrNumericList, temperature_units: str,
                    interpolator_type: str, **kwargs,
                    ) -> np.ndarray:
        """Interpolates and/or extrapolates fluid property data for given
        pressures and temperatures

        Retrieves fluid property data from the file for given pressure(s) and
        temperature(s), interpolating and/or extrapolating if necessary.
        Interpolation is performed with the :py:mod:`scipy.interpolate` package
        (https://docs.scipy.org/doc/scipy/reference/interpolate.html).

        Parameters
        ----------
        fluid_property : str
            The name of the fluid property for which data should be returned
        output_units : str
            The units with which to return fluid property data
        pressures : float or np.ndarray or list or tuple
            The pressure(s) at which fluid properties should be returned.  See
            the "Notes" section for more information about required format
        pressure_units : str
            The units in which the ``pressures`` argument is provided
        temperatures : float or np.ndarray or list or tuple
            The temperature(s) at which fluid properties should be returned.
            See the "Notes" section for more information about required format
        temperature_units : str
            The units in which the ``temperatures`` argument is provided
        interpolator_type : str
            The SciPy interpolation function to use to perform interpolation.
            Can be selected from any of the options in the "Notes" section
        **kwargs
            Any keyword arguments to be supplied to the SciPy interpolation
            function specified by ``interpolator_type``.  See the "Notes"
            section for more information

        Returns
        -------
        np.ndarray
            The interpolated value(s) of the fluid properties given by
            ``fluid_property`` for the pressures and temperatures given
            by ``pressures`` and ``temperatures``, respectively.  See the
            "Notes" section for more information about the output data format

        Notes
        -----
        **Valid Fluid Properties**

        The table below summarizes valid inputs for the ``fluid_property``
        argument.

        .. list-table::
            :align: left
            :header-rows: 1
            :widths: auto

            * - Fluid Property
              - Valid ``fluid_property`` Inputs
              - Sample Units
            * - Density
              - ``rho``, ``density``
              - ``kg/m^3``
            * - Bulk modulus
              - ``k``, ``bulk modulus``
              - ``Pa_a``
            * - Kinematic viscosity
              - ``nu``, ``kinematic viscosity``
              - ``m^2/s``
            * - Specific heat capacity
              - ``cp``, ``specific heat capacity``
              - ``J/kg/K``
            * - Thermal conductivity
              - ``lambda``, ``thermal conductivity``
              - ``W/m/K``
            * - Volumetric expansion coefficient
              - ``alpha``, ``volumetric expansion coefficient``
              - ``1/K``
            * - Specific enthalpy
              - ``h``, ``specific enthalpy``
              - ``J/kg``
            * - Dynamic viscosity
              - ``mu``, ``dynamic viscosity``, ``absolute viscosity``
              - ``Pa_a*s``

        **Interpolation Functions**

        The following interpolation functions are available (set the
        ``interpolator_type`` argument to the given string to use each):

        - ``'interpn'``: The :py:class:`scipy.interpolate.interpn` function
          for multidimensional interpolation on rectangular grids (`reference
          <https://docs.scipy.org/doc/scipy/reference/generated
          /scipy.interpolate.interpn.html>`__)
        - ``'griddata'``: The :py:class:`scipy.interpolate.griddata`
          interpolation function for unstructured, multivariate interpolation
          (`reference <https://docs.scipy.org/doc/scipy/reference/generated
          /scipy.interpolate.griddata.html>`__)
        - ``'RBFInterpolator'``: The :py:class:`scipy.interpolate.RBFInterpolator`
          function for unstructured, multivariate interpolation using a radial
          basis function (`reference <https://docs.scipy.org/doc/scipy/reference
          /generated/scipy.interpolate.RBFInterpolator.html>`__)
        - ``'CloughTocher2DInterpolator'``: The
          :py:class:`scipy.interpolate.CloughTocher2DInterpolator`
          function for unstructured, multivariate interpolation using a
          piecewise cubic, :math:`C^1` smooth, curvature-minimizing interpolant
          (`reference <https://docs.scipy.org/doc/scipy/reference/generated
          /scipy.interpolate.CloughTocher2DInterpolator.html>`__)

        Review the SciPy documentation for questions about parameters for the
        interpolation functions.  *Some of these interpolation functions
        require users to set* ``kwargs`` *to meet specific mathematical
        requirements, and errors may be encountered if such requirements are
        not met.*

        **Formatting of Pressure and Temperature Inputs**

        Interpolation can be performed for one or more pressures and/or
        temperatures with a single call of the :py:meth:`interpolate` method.
        The table below specifies the valid format of the ``pressures`` and
        ``temperatures`` arguments, and the corresponding format of data that
        would be returned by each selection of inputs.

        .. list-table::
            :align: left
            :header-rows: 1
            :widths: auto

            * - Input: ``pressures``
              - Input: ``temperatures``
              - Output Shape
              - Notes
            * - ``float``
              - ``float``
              - ``(1,)``
              - Returns the interpolated property at the input pressure and
                temperature
            * - ``list`` (length ``n``)
              - ``float``
              - ``(n,)``
              - Returns the interpolated property at every value of
                ``pressures`` for the specified temperature
            * - ``float``
              - ``list`` (length ``n``)
              - ``(n,)``
              - Returns the interpolated property at every value of
                ``temperatures`` for the specified pressure
            * - ``list`` (length ``n``)
              - ``list`` (length ``n``)
              - ``(n,)``
              - Input lists **must** have equal length.  Component ``i`` of
                the returned array is interpolated for ``pressures[i]`` and
                ``temperatures[i]``

        Note that a length-1 list is considered equivalent to ``float`` in the
        table above.  Also, NumPy arrays or tuples can be used in place of
        ``list`` above.
        """
        fluid_property = str(fluid_property).lower()

        for _, _, _, attr in self.__file_format:
            if getattr(self, attr) is None:
                raise FileNotParsedError(
                    'Unable to interpolate fluid properties; file has not '
                    'been read/parsed')

        if (self._pressure_values is None) or (self._temperature_values is None):
            raise FileNotParsedError(
                'Unable to interpolate fluid properties; file has not '
                'been read/parsed')

        # Convert pressures and temperatures to NumPy arrays
        pressures = to_np_1D_array(pressures, dtype=np.float64)
        temperatures = to_np_1D_array(temperatures, dtype=np.float64)

        num_pressures = len(pressures)
        num_temperatures = len(temperatures)

        # Check that input array sizes are compatible
        if (
            (num_pressures > 1) and (num_temperatures > 1)
            and (num_pressures != num_temperatures)
        ):
            raise ValueError(
                'Pressure and temperature arrays both have lengths greater '
                'than 1, but their lengths are not equal')

        # Ensure pressure and temperature arrays have the same size
        if (num_pressures > 1) and (num_temperatures == 1):
            temperatures = np.repeat(temperatures, num_pressures)

        elif ((num_temperatures > 1) and (num_pressures == 1)):
            pressures = np.repeat(pressures, num_temperatures)

        # Retrieve data to interpolate/extrapolate
        data: np.ndarray = np.array([])
        data_units = ''

        success = False
        for symbol, desc, units, attr in self.__file_format:
            if fluid_property in (symbol, desc):
                data = getattr(self, attr)
                data_units = units

                success = True
                break

        if not success:
            if fluid_property in ('mu', 'absolute viscosity',
                                  'dynamic viscosity'):
                data = np.multiply(self._density, self._viscosity_k)  # type: ignore
                data_units \
                    = f'{self.__file_format[0][2]}*{self.__file_format[2][2]}'

                success = True

        if not success:
            raise ValueError(f'Invalid fluid property "{fluid_property}"')

        # Perform interpolation/extrapolation
        pressures = self.unit_converter.convert(
            quantity  = pressures,
            from_unit = pressure_units,
            to_unit   = self.__pressure_units
        )
        temperatures = self.unit_converter.convert(
            quantity  = temperatures,
            from_unit = temperature_units,
            to_unit   = self.__temperature_units
        )

        if interpolator_type == 'interpn':
            output_data = scipy.interpolate.interpn(
                points = (self._temperature_values, self._pressure_values),
                values = data,
                xi     = np.transpose(np.array([temperatures, pressures])),
                **kwargs
            )

        elif interpolator_type in ('griddata', 'RBFInterpolator',
                                   'CloughTocher2DInterpolator'):
            mesh_p, mesh_t = np.meshgrid(self._pressure_values,
                                         self._temperature_values)
            mesh_points = np.transpose(np.array([mesh_t.flatten(),
                                                 mesh_p.flatten()]))
            data_flat = data.flatten()

            query_points = np.transpose(np.array([temperatures, pressures]))

            if interpolator_type == 'griddata':
                output_data = scipy.interpolate.griddata(
                    points = mesh_points,
                    values = data_flat,
                    xi     = query_points,
                    **kwargs
                )

            elif interpolator_type == 'RBFInterpolator':
                interpolator = scipy.interpolate.RBFInterpolator(
                    y = mesh_points,
                    d = data_flat,
                    **kwargs
                )

                output_data = interpolator(query_points)

            else:
                interpolator = scipy.interpolate.CloughTocher2DInterpolator(
                    points = mesh_points,
                    values = data_flat,
                    **kwargs
                )

                output_data = interpolator(query_points)

        else:
            raise ValueError(
                f'Interpolator "{interpolator_type}" is not supported')

        return self.unit_converter.convert(
            quantity  = output_data,
            from_unit = data_units,
            to_unit   = output_units,
        )

    def parse(self) -> None:
        """Parses the file content in the :py:attr:`contents` list and
        populates attributes (such as :py:attr:`title`) with extracted data

        This method parses the data in :py:attr:`contents`, extracting
        fluid property data and storing it in this object's attributes for
        easier reading and editing.
        """
        super().parse()
        original_contents = copy.deepcopy(self.contents)

        # Settings
        tolerance = 1e-8  # tolerance for equality of floating-point numbers

        # Remove comments from file
        self.clean_contents(
            remove_comments    = True,
            strip              = True,
            remove_blank_lines = True,
        )

        # READ PRESSURE AND TEMPERATURE METADATA
        if len(self.contents) < 6:
            raise FluidPropertyFileError(
                'Fluid property file is missing required lines listing '
                'temperature and pressure ranges')

        try:
            # Get number of temperature values and temperature step
            self._num_temperature = int(float(self.contents.pop(0)))
            self._step_temperature = float(self.contents.pop(0))

            # Get number of pressure values and pressure step
            self._num_pressure = int(float(self.contents.pop(0)))
            self._step_pressure = float(self.contents.pop(0))

            # Get list of all temperature and pressure values
            self._temperature_values = np.array(
                [float(i) for i in self.contents.pop(0).split()],
                dtype=np.float64,
            )
            self._pressure_values = np.array(
                [float(i) for i in self.contents.pop(0).split()],
                dtype=np.float64,
            )
        except ValueError as exception:
            raise FluidPropertyFileError(
                'Fluid property file is missing required temperature and '
                'pressure metadata') from exception

        # VALIDATE FILE FORMAT
        # Check that file has expected number of lines
        expected_len = self.num_pressure * self.num_temperature
        if len(self.contents) != expected_len:
            raise FluidPropertyFileError(
                f'Fluid property file has {len(self.contents)} lines of fluid '
                f'property data, but {expected_len} lines were expected')

        # Check that temperature and pressure steps match listed values
        max_diff = np.max(np.abs(
            (self._temperature_values[1:] - self._temperature_values[:-1])
            - self._step_temperature
        ))
        if max_diff > tolerance:
            raise FluidPropertyFileError(
                'Fluid property file temperature values do not have stated '
                f'temperature step (maximum difference: {max_diff})')

        max_diff = np.max(np.abs(
            (self._pressure_values[1:] - self._pressure_values[:-1])
            - self._step_pressure
        ))
        if max_diff > tolerance:
            raise FluidPropertyFileError(
                'Fluid property file pressure values do not have stated '
                f'pressure step (maximum difference: {max_diff})')

        # READ FLUID PROPERTY DATA
        num_samples = self.num_pressure * self.num_temperature

        # Read data from file
        try:
            raw_file_data = np.transpose(np.array(
                [line.strip().split() for line in self.contents],
                dtype=np.float64,
            ))
        except ValueError as exception:
            raise FluidPropertyFileError(
                'Unable to read fluid properties file: some data in the file '
                'are not numeric') from exception

        if raw_file_data.shape != (7, num_samples):
            raise FluidPropertyFileError(
                'Unable to read fluid properties file: data array has shape '
                f'{raw_file_data.shape}, but expected {(7, num_samples)}')

        # Store data -- columns correspond to different pressures, rows
        # correspond to different temperatures
        shape = (self.num_temperature, self.num_pressure)

        for i, prop in enumerate(self.__file_format):
            attr_name = prop[3]
            setattr(self, attr_name, np.reshape(raw_file_data[i], shape))

        self.contents.clear()
        self.contents.extend(original_contents)



"""Default settings and templates for creating plots for the
:py:mod:`mahautils.shapes` module.
"""

from typing import Optional

# Mypy type checking disabled for packages that are not PEP 561-compliant
import plotly.graph_objects as go  # type: ignore

_figure_config = {
    'edits': {
        'annotationPosition': True,
        'annotationTail': True,
        'annotationText': True,
        'legendPosition': True,
    },
    'modeBarButtonsToAdd': [
        'drawclosedpath',
        'drawopenpath',
        'drawline',
        'drawrect',
        'drawcircle',
        'eraseshape',
    ],
    'toImageButtonOptions': {
        'format': 'png',
        'width': None,
        'height': None,
    },
    'scrollZoom': True,
    'showAxisDragHandles': True,
}


def _create_blank_plotly_figure(units: Optional[str] = None):
    """Creates a blank Plotly figure with general formatting options, intended
    to be used to plot :py:class:`mahautils.shapes.Layer` and
    :py:class:`mahautils.shapes.Canvas` objects

    Parameters
    ----------
    units : str, optional
        Units to display in axis titles, or ``None`` to suppress showing units
        in axis titles (default is ``None``)
    """
    figure = go.Figure()

    # Axis titles
    if units is None:
        figure.update_layout(xaxis_title='x')
        figure.update_layout(yaxis_title='y')
    else:
        figure.update_layout(xaxis_title=f'x [{units}]')
        figure.update_layout(yaxis_title=f'y [{units}]')

    # Use equal scale for x- and y-axes
    figure.update_yaxes(scaleanchor='x', scaleratio=1)

    # Default format settings
    figure.update_layout(margin={'t': 30, 'r': 20})

    default_axes_settings = {
        # Border around plot
        'showline': True, 'linewidth': 1, 'linecolor': 'black', 'mirror': True,

        # Settings for x=0 and y=0 axes
        'zeroline': True, 'zerolinewidth': 1, 'zerolinecolor': '#7b7b7b',

        # Settings for gridlines
        'showgrid': True, 'gridwidth': 1, 'gridcolor': '#d2d2d2',
    }
    figure.update_xaxes(**default_axes_settings)
    figure.update_yaxes(**default_axes_settings)

    figure.update_layout(plot_bgcolor='white')

    # Hide legend
    figure.update_layout(showlegend=False)

    return figure



"""This module provides a customized dictionary which can be used to store
data in a key-value format.
"""

from typing import List, Optional, OrderedDict, Tuple, Type, TypeVar, Union

from pyxx.arrays import max_list_item_len

K = TypeVar('K')
V = TypeVar('V')


class Dictionary(OrderedDict[K, V]):
    """Customized Python dictionary

    This class is a modified version of Python's built-in :py:class:`OrderedDict`
    dictionary.  Most typical dictionary methods function as they do in a
    :py:class:`dict`.  However, additional functionality has been added, such as
    being able to insert items at any position in the dictionary and
    customizations for printing content in the dictionary.
    """

    def __init__(self, contents: Optional[dict] = None,
                 required_key_type: Optional[
                     Union[Type[K], List[Type[K]], Tuple[Type[K], ...]]] = None,
                 required_value_type: Optional[
                     Union[Type[V], List[Type[V]], Tuple[Type[V], ...]]] = None,
                 multiline_print: bool = False,
                 str_indent: int = 0,
                 str_pad_left: int = 1, str_pad_right: int = 2,
                 custom_except_class: Type[Exception] = KeyError,
                 custom_except_msg: str = '%s',
                 ) -> None:
        """Defines a :py:class:`Dictionary` instance

        Defines a new :py:class:`Dictionary` object, providing options to
        populate the dictionary's contents and define parameters controlling
        how the dictionary is formatted when converting to a string.

        Parameters
        ----------
        contents : dict, optional
            Python ``dict`` which, if provided, will immediately populate the
            data in the dictionary upon creation (default is ``None``)
        required_key_type : type or list or tuple, optional
            One or more required type(s) of which all keys in the dictionary
            must be a subclass.  Set to ``None`` to allow any type (default is
            ``None``)
        required_value_type : type or list or tuple, optional
            One or more required type(s) of which all values in the dictionary
            must be a subclass.  Set to ``None`` to allow any type (default is
            ``None``)
        multiline_print : bool, optional
            Whether to print each entry of the dictionary on its own line when
            converting to a printable string representation (default is
            ``False``)
        str_indent : int, optional
            Sets the dictionary's :py:attr:`str_indent` attribute
            (default is ``0``).  Only applicable if ``multiline_print`` is
            ``True``
        str_pad_left : int, optional
            Sets the dictionary's :py:attr:`str_pad_left` attribute (default
            is ``0``).  Only applicable if ``multiline_print`` is ``True``
        str_pad_right : int, optional
            Sets the dictionary's :py:attr:`str_pad_right` attribute (default
            is ``0``).  Only applicable if ``multiline_print`` is ``True``
        custom_except_class : class, optional
            An :py:class:`Exception` subclass to raise if a given key is not
            found in the dictionary (default is :py:class:`KeyError`)
        custom_except_msg : str, optional
            Only applicable if ``custom_except_class`` is not ``None``.
            Specifies an error message to throw if a key is not found in the
            dictionary.  Must contain a single ``%s`` occurrence, which will
            be replaced by the name of the key which was not found (default
            is ``'%s'``)
        """
        super().__init__()

        # Store required key and value types
        self._required_key_type = required_key_type
        self._required_value_type = required_value_type

        # Store parameters for representing dictionary as string
        self.multiline_print = multiline_print
        self.str_indent = str_indent
        self.str_pad_left = str_pad_left
        self.str_pad_right = str_pad_right

        # Store customized exception information
        self.custom_except_class = custom_except_class
        self.custom_except_msg = custom_except_msg

        # Store dictionary contents if provided
        if contents is not None:
            self.update(contents)

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            raise self.custom_except_class(  # pylint: disable=W0707
                self.custom_except_msg % key)

    def __repr__(self) -> str:
        if not self._multiline_print:
            return super().__repr__()

        return str(self)

    def __setitem__(self, key, value):
        if (
            self._required_key_type is not None
            and not isinstance(key, self._required_key_type)
        ):
            raise TypeError(
                'Dictionary keys must be of one of the following '
                f'types: {self._required_key_type}')

        if (
            self._required_value_type is not None
            and not isinstance(value, self._required_value_type)
        ):
            raise TypeError(
                'Dictionary values must be of one of the following '
                f'types: {self._required_value_type}')

        super().__setitem__(key, value)

    def __str__(self) -> str:
        if not self._multiline_print:
            return super().__str__()

        if len(self) > 0:
            # Determine maximum string length of keys in dictionary
            _max_key_len = max_list_item_len(self.keys())

            # Create dictionary string representation
            representation = '\n'.join(
                [(f'{" "*self.str_indent}{key:{_max_key_len+self.str_pad_left}s}'
                  f':{" "*self.str_pad_right}{value}')
                 for key, value in self.items()])
        else:
            representation = ''

        return representation

    @property
    def custom_except_class(self) -> Type[Exception]:
        """The exception to raise if a key does not exist in the dictionary"""
        return self._custom_except_class

    @custom_except_class.setter
    def custom_except_class(self, custom_except_class: Type[Exception]) -> None:
        if not issubclass(custom_except_class, Exception):
            raise TypeError(
                'Custom exceptions must be a subclass of `Exception`')

        self._custom_except_class = custom_except_class

    @property
    def custom_except_msg(self) -> str:
        """A custom error message if a key was not found in the dictionary

        The message must contain a single ``%s`` occurrence, which will
        be replaced by the name of the key which was not found.
        """
        return self._custom_except_msg

    @custom_except_msg.setter
    def custom_except_msg(self, custom_except_msg: str) -> None:
        if not (
            custom_except_msg.find('%s')
            == custom_except_msg.rfind('%s')
            != -1
        ):
            raise ValueError('Custom error message must contain exactly one '
                             'occurrence of "%s"')

        self._custom_except_msg = str(custom_except_msg)

    @property
    def multiline_print(self) -> bool:
        """Whether to print each entry of the dictionary on its own line when
        converting to a printable string representation"""
        return self._multiline_print

    @multiline_print.setter
    def multiline_print(self, multiline_print: bool) -> None:
        self._multiline_print = bool(multiline_print)

    @property
    def str_indent(self) -> int:
        """The number of spaces of indentation at the beginning of each line
        when converting a :py:class:`Dictionary` object to a printable
        string representation.  Only applicable if :py:attr:`multiline_print`
        is ``True``"""
        return self._str_indent

    @str_indent.setter
    def str_indent(self, str_indent: int):
        self._str_indent = int(str_indent)

    @property
    def str_pad_left(self) -> int:
        """The number of spaces between the printed key and the separator when
        converting a :py:class:`Dictionary` object to a printable string
        representation.  Only applicable if :py:attr:`multiline_print` is
        ``True``"""
        return self._str_pad_left

    @str_pad_left.setter
    def str_pad_left(self, str_pad_left: int):
        self._str_pad_left = int(str_pad_left)

    @property
    def str_pad_right(self) -> int:
        """The number of spaces between the separator and printed value when
        converting a :py:class:`Dictionary` object to a printable string
        representation.  Only applicable if :py:attr:`multiline_print` is
        ``True``"""
        return self._str_pad_right

    @str_pad_right.setter
    def str_pad_right(self, str_pad_right: int):
        self._str_pad_right = int(str_pad_right)

    def delete_index(self, index: int) -> None:
        """Removes an item from the dictionary based on its index

        Items can easily be removed from the dictionary by key (example: to
        remove an item with key ``'myKey'`` from a dictionary, simply call
        ``del dictionary['myKey']``).  This method extends this functionality
        and allows items to be removed based on their position in the
        dictionary.

        Parameters
        ----------
        index : int
            The index of the item to remove from the dictionary
        """
        del self[list(self.keys())[index]]

    def index(self, key: K) -> int:
        """Returns the index of a key in the dictionary

        Parameters
        ----------
        key : Any
            A key in the dictionary

        Returns
        -------
        int
            The location (beginning from 0) of ``key`` in the dictionary
        """
        if key not in self:
            raise KeyError(f'Key "{key}" not present in dictionary')

        return list(self.keys()).index(key)

    def insert(self, index: int, key: K, value: V) -> None:
        """Adds a new item to the dictionary at a specified position

        Parameters
        ----------
        index : int
            The position in the dictionary at which to insert the new key
        key : Any
            The key for the new item to insert int the dictionary
        value : Any
            The value for the new item to insert in the dictionary
        """
        if key in self:
            raise KeyError(f'Key "{key}" already exists, unable to overwrite')

        keys = list(self.keys())

        # Insert at beginning of dictionary
        self[key] = value

        # Rearrange items in dictionary so that new item is in the
        # correct position
        if index < len(self) / 2:
            self.move_to_end(key, last=False)

            for k in reversed(keys[0:index]):
                self.move_to_end(k, last=False)
        else:
            for k in reversed(keys[index:]):
                self.move_to_end(k, last=True)

        # It is not optimal to rearrange items iteratively, but the parent
        # `OrderedDict` class is written in C so we don't have access to
        # internal attributes and pointers that would make it faster (it would
        # be possible to rewrite the MahaUtils `Dictionary` for improved
        # efficiency, but the intended applications of the package
        # don't necessitate fast performance)

    def insert_after(self, existing_key: K, key: K, value: V) -> None:
        """Adds a new item to the dictionary at after another existing item
        in the dictionary

        Parameters
        ----------
        existing_key : Any
            The existing key in the dictionary after which ``key`` should
            be inserted
        key : Any
            The key for the new item to insert int the dictionary
        value : Any
            The value for the new item to insert in the dictionary
        """
        self.insert(self.index(existing_key) + 1, key, value)

    def insert_before(self, existing_key: K, key: K, value: V) -> None:
        """Adds a new item to the dictionary at before another existing item
        in the dictionary

        Parameters
        ----------
        existing_key : Any
            The existing key in the dictionary before which ``key`` should
            be inserted
        key : Any
            The key for the new item to insert int the dictionary
        value : Any
            The value for the new item to insert in the dictionary
        """
        self.insert(self.index(existing_key), key, value)



"""Packages to support research work at the Maha Fluid Power Research Center

This package contains a number of utilities that are useful for performing
research work and interacting with the Maha Multics software.
"""


# PROGRAM VERSION ------------------------------------------------------------
__version__ = '0.2.0'


# PACKAGE MODULES ------------------------------------------------------------
from . import multics
from . import shapes
from . import utils



"""Code to set up configuration panel of the simulation results viewer.
"""

from .utils_generate_table import generate_file_table_body
from .main_panel import simviewer_config_panel
from .settings_general import render_general_settings
from .settings_x import render_x_settings
from .settings_y import render_y_settings



"""Panel tab for configuring general plot settings.
"""

# Mypy type checking disabled for packages that are not PEP 561-compliant
import dash                              # type: ignore
import dash_bootstrap_components as dbc  # type: ignore

from mahautils.multics.sim_results_viewer.constants import (
    TAB_BAR_PADDING,
    UI_DESCRIPTION_MARGIN_BELOW,
)
from mahautils.multics.sim_results_viewer.store import default_plot_config_general
from mahautils.multics.sim_results_viewer.ui_elements import (
    color_picker,
    on_off_switch_row,
)


import_export_button_styling = {
    'width': '100%',
    'textAlign': 'center',
    'fontWeight': 'bold',
}


def plot_settings_general():
    """Creates the configuration panel tab with general plot settings"""
    return [
        dash.html.H4('Load/Save Plot Configuration',
                     style={'marginTop': TAB_BAR_PADDING},),
        dash.html.Hr(),
        dbc.Row([
            dbc.Col(
                dbc.Spinner(
                    dash.dcc.Upload(
                        dbc.Button('Import', id='config-import-button',
                                   style=import_export_button_styling),
                        id='plot-config-upload',
                    ),
                    # Delay prevents showing spinner when changing other
                    # configuration UI elements
                    delay_show=200,
                ),
                class_name='d-grid gap-2',
            ),
            dbc.Col(
                [
                    dbc.Button('Export', id='config-export-button',
                               style=import_export_button_styling),
                    dash.dcc.Download(id='config-export-download'),
                ],
                class_name='d-grid gap-2',
            ),
        ]),
        dash.html.Br(),
        dash.html.Br(),
        dash.html.Div(
            render_general_settings(default_plot_config_general),
            id='plot-config-general-settings'
        ),
    ]


def render_general_settings(config_general: dict):
    """Generates the UI elements that allow the user to configure general
    plot settings"""
    return [
        dash.html.H4('Plot Appearance'),
        dash.html.Hr(),
        dash.html.H5('Plot Title'),
        dbc.Input(
            debounce=True,
            id={'component': 'plot-config', 'tab': 'general',
                'field': 'title'},
            placeholder='Enter a title for the plot...',
            value=config_general['title'],
        ),
        dash.html.Br(),
        dash.html.H5('Background Color'),
        color_picker(
            identifier={'component': 'plot-config', 'tab': 'general',
                        'field': 'background-color'},
            current_color=config_general['background_color'],
        ),
        dash.html.Br(),
        dash.html.H5('Gridlines'),
        on_off_switch_row(
            identifier={'component': 'plot-config', 'tab': 'general',
                        'field': 'grid-x'},
            state=config_general['grid_x'],
            description='Enable x-axis gridlines'),
        on_off_switch_row(
            identifier={'component': 'plot-config', 'tab': 'general',
                        'field': 'grid-y'},
            state=config_general['grid_y'],
            description='Enable y-axis gridlines'),
        dash.html.Br(),
        dash.html.H5('Units'),
        on_off_switch_row(
            identifier={'component': 'plot-config', 'tab': 'general',
                        'field': 'append-units'},
            state=config_general['append_units'],
            description=('Append units to axis titles and legend entries '
                         '(where applicable)')),
        dash.html.Br(),
        dash.html.H5('Vertical Axis Spacing'),
        dash.html.P(
            ('Sets the width in the plot window allocated for each vertical '
             'y-axis.  Must be a number in the range [0, 1]'),
            style={'marginBottom': UI_DESCRIPTION_MARGIN_BELOW},
        ),
        dbc.Input(
            debounce=True,
            value=config_general['width_per_y_axis'],
            id={'component': 'plot-config', 'tab': 'general',
                'field': 'width-per-y-axis'},
            style={'width': 200},
        ),
        dash.html.Br(),
        dash.html.Br(),
        dash.html.H4('User Interface'),
        dash.html.Hr(),
        dash.html.H5('Hover Labels'),
        dash.html.P(
            ('Controls the information displayed when hovering the cursor '
             'over the plot'),
            style={'marginBottom': UI_DESCRIPTION_MARGIN_BELOW},
        ),
        dash.dcc.Dropdown(
            options=['closest', 'x unified', 'y unified', 'x', 'y'],
            value=config_general['hovermode'],
            id={'component': 'plot-config', 'tab': 'general',
                'field': 'hovermode'},
            clearable=False,
            multi=False,
        ),
        dash.html.Br(),
        dash.html.H5('Plot Updates'),
        on_off_switch_row(
            identifier={'component': 'plot-config', 'tab': 'general',
                        'field': 'freeze-uirevision'},
            state=config_general['freeze_uirevision'],
            description='Prevent user interface state reset on plot updates'),
        dbc.Tooltip(
            ('Prevents changes to plot data or settings from resetting past '
             'user interactions (zoom, clicking legend items, shape '
             'annotations, etc.)'),
            target={'component': 'plot-config', 'tab': 'general',
                    'field': 'freeze-uirevision'},
            trigger='hover',
        ),
        dash.html.Br(),
        dash.html.Br(),
        dash.html.H4('Image Export Options'),
        dash.html.Hr(),
        dash.html.H5('Export Type'),
        dash.dcc.Dropdown(
            options=['png', 'jpeg', 'webp', 'svg'],
            value=config_general['image_export_type'],
            id={'component': 'plot-config', 'tab': 'general',
                'field': 'image_export_type'},
            clearable=False,
            multi=False,
        ),
        dash.html.Br(),
        dash.html.H5('Aspect Ratio'),
        dash.html.P(
            ('Output aspect ratio for exported images (width x height). '
             'Leave blank to use window size'),
            style={'marginBottom': UI_DESCRIPTION_MARGIN_BELOW},
        ),
        dash.html.Div([
            dbc.Input(
                debounce=True,
                id={'component': 'plot-config', 'tab': 'general',
                    'field': 'image_export_width'},
                value=config_general['image_export_width'],
                style={'display': 'inline-block', 'width': '100px'},
            ),
            dash.html.P('  x  ', style={'display': 'inline-block',
                                        'whiteSpace': 'pre'}),
            dbc.Input(
                debounce=True,
                id={'component': 'plot-config', 'tab': 'general',
                    'field': 'image_export_height'},
                value=config_general['image_export_height'],
                style={'display': 'inline-block', 'width': '100px'},
            ),
        ]),
        dash.html.Br(),
        dash.html.H5('Resolution'),
        dash.html.P(
            ('Setting this to n will increase the exported image '
             'resolution by a factor of n'),
            style={'marginBottom': UI_DESCRIPTION_MARGIN_BELOW},
        ),
        dbc.Input(
            debounce=True,
            id={'component': 'plot-config', 'tab': 'general',
                'field': 'image_export_scale'},
            value=config_general['image_export_scale'],
            style={'width': '100px'},
        ),
    ]



"""Unit conversion utilities for the :py:mod:`mahautils.multics` module.  The
classes in this module are intended to aid in performing unit conversions when
reading and writing Maha Multics-related files.  The units typically defined
in Maha Multics input files are included by default in these unit converter
objects.
"""

import math
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pyxx

from .constants import (
    _DEG_C_TO_K_OFFSET,
    _DEG_F_TO_K_OFFSET,
    _DEG_F_TO_K_SCALE,
    _LBF_TO_N_SCALE,
    _LBM_TO_KG_SCALE,
    _IN_TO_M_SCALE,
    _PSI_TO_PA_SCALE,
    ATMOSPHERIC_PRESSURE_PA,
)


# Custom system of units
class MahaMulticsUnitSystem(pyxx.units.UnitSystem):
    """A Maha Multics-specific system of units

    This class is a modified version of the :py:class:`pyxx.units.UnitSystem`
    class.  It represents a system of units similar to SI, with the following
    base units:

    1. Mass: kilogram [:math:`kg`]
    2. Length: meter [:math:`m`]
    3. Time: second [:math:`s`]
    4. Temperature: Kelvin [:math:`K`]
    5. Amount of substance: mole [:math:`mol`]
    6. Electric current: ampere [:math:`A`]
    7. Luminous intensity: candela [:math:`cd`]

    By using a custom system of units for Maha Multics, it automatically
    ensures that unit conversions can only be performed for Maha
    Multics-compatible units.
    """

    def __init__(self):
        """Creates an instance of a Maha Multics-specific system of units
        with seven base units
        """
        super().__init__(
            num_base_units=7,
            name='SI',
            description=('International System of Units, customized for the '
                         'Maha Multics software')
        )


# Custom unit class that automatically uses the Maha Multics-specific
# system of units
class MahaMulticsUnit(pyxx.units.UnitLinear):
    """Class for representing units with linear transformations to/from
    the base units, customized to the Maha Multics software

    Defines a unit in which the transformations to/from the base units
    of the system of units (that is, the functions given by
    :py:attr:`from_base_function` and :py:attr:`to_base_function`) are linear.
    A large portion of the units encountered in everyday use can be considered
    linear units, so this class was created to simplify defining such units.

    Notes
    -----
    To convert an array of arbitrary dimensions ``inputs`` from the given
    object's unit to the base units, the following equation is applied:

    .. code-block:: python

        outputs = (scale * inputs) + offset
    """

    def __init__(self,
                 base_unit_exps:
                     Union[List[float], Tuple[float, ...], np.ndarray],
                 scale: float, offset: float,
                 identifier: Optional[str] = None, name: Optional[str] = None
                 ) -> None:
        """Creates an instance of the :py:class:`MahaMulticsUnit` class

        Defines an object representing a base or derived unit in which the
        functions converting a value to/from the base units of the system
        of units :py:attr:`unit_system` are linear functions and the system
        of units is specific to the Maha Multics software.

        Parameters
        ----------
        base_unit_exps : list or tuple or np.ndarray
            A 1D list of exponents relating the given object's unit to the
            base units of ``unit system``
        scale : float
            The multiplicative factor applied when converting from
            the given object's unit to the base units
        offset : float
            The constant value added when converting from the given
            object's unit to the base units
        identifier : str, optional
            A short identifier describing the unit (example: ``'kg'``)
            (default is ``None``)
        name : str, optional
            A name describing the unit (example: ``'kilogram'``) (default
            is ``None``)
        """
        super().__init__(
            unit_system    = MahaMulticsUnitSystem(),
            base_unit_exps = base_unit_exps,
            scale          = scale,
            offset         = offset,
            identifier     = identifier,
            name           = name
        )


_MAHA_MULTICS_DEFAULT_UNITS: Dict[str, Dict[str, Any]] = {
    ## BASE UNITS ------------------------------------------------------------
    'kg': {
        'unit': MahaMulticsUnit((1, 0, 0, 0, 0, 0, 0), scale=1, offset=0),
        'tags': ('mass',),
        'name': 'kilogram',
        'description': 'Unit of measure of mass',
        'aliases': ('kilogram', 'kilograms'),
    },
    'm': {
        'unit': MahaMulticsUnit((0, 1, 0, 0, 0, 0, 0), scale=1, offset=0),
        'tags': ('length',),
        'name': 'meter',
        'aliases': ('meter', 'meters'),
    },
    's': {
        'unit': MahaMulticsUnit((0, 0, 1, 0, 0, 0, 0), scale=1, offset=0),
        'tags': ('time',),
        'name': 'second',
        'aliases': ('sec', 'second', 'seconds'),
    },
    'K': {
        'unit': MahaMulticsUnit((0, 0, 0, 1, 0, 0, 0), scale=1, offset=0),
        'tags': ('temperature',),
        'name': 'Kelvin',
    },
    'mol': {
        'unit': MahaMulticsUnit((0, 0, 0, 0, 1, 0, 0), scale=1, offset=0),
        'name': 'mole',
        'aliases': ('moles', 'mole'),
    },
    'A': {
        'unit': MahaMulticsUnit((0, 0, 0, 0, 0, 1, 0), scale=1, offset=0),
        'name': 'ampere',
        'description': 'Unit of measure of electric current',
        'aliases': ('amp', 'amps'),
    },
    'cd': {
        'unit': MahaMulticsUnit((0, 0, 0, 0, 0, 0, 1), scale=1, offset=0),
        'tags': ('luminance',),
        'name': 'candela',
    },

    ## DIMENSIONLESS QUANTITIES ----------------------------------------------
    'dimensionless': {
        'unit': MahaMulticsUnit((0, 0, 0, 0, 0, 0, 0), scale=1, offset=0),
        'tags': ('dimensionless',),
        'name': 'dimensionless',
        'description': 'Unit assigned to dimensionless numbers',
        'aliases': ('-', 'unitless', 'null',),
    },

    ## PERCENTAGES -----------------------------------------------------------
    '%': {
        'unit': MahaMulticsUnit((0, 0, 0, 0, 0, 0, 0), scale=0.01, offset=0),
        'tags': ('percentage',),
        'name': 'percentage',
    },

    ## ANGLES ----------------------------------------------------------------
    'rad': {
        'unit': MahaMulticsUnit((0, 0, 0, 0, 0, 0, 0), scale=1, offset=0),
        'tags': ('angle',),
        'name': 'radian',
        'aliases': ('radian', 'radians'),
    },
    'deg': {
        'unit': MahaMulticsUnit((0, 0, 0, 0, 0, 0, 0), scale=math.pi/180,
                                offset=0),
        'tags': ('angle',),
        'name': 'degrees',
        'aliases': ('degree', 'degrees'),
    },
    'rev': {
        'unit': MahaMulticsUnit((0, 0, 0, 0, 0, 0, 0), scale=2*math.pi,
                                offset=0),
        'tags': ('angle',),
        'name': 'revolutions',
    },

    ## LENGTH ----------------------------------------------------------------
    'mm': {
        'unit': MahaMulticsUnit((0, 1, 0, 0, 0, 0, 0), scale=0.001, offset=0),
        'tags': ('length',),
        'name': 'millimeter',
        'aliases': ('millimeter', 'millimeters'),
    },
    'cm': {
        'unit': MahaMulticsUnit((0, 1, 0, 0, 0, 0, 0), scale=0.01, offset=0),
        'tags': ('length',),
        'name': 'centimeter',
        'aliases': ('centimeter', 'centimeters'),
    },
    'micron': {
        'unit': MahaMulticsUnit((0, 1, 0, 0, 0, 0, 0), scale=1e-6, offset=0),
        'tags': ('length',),
        'name': 'micron',
        'aliases': ('microns',),
    },
    'in': {
        'unit': MahaMulticsUnit((0, 1, 0, 0, 0, 0, 0), scale=_IN_TO_M_SCALE,
                                offset=0),
        'tags': ('length',),
        'name': 'inch',
        'aliases': ('inch', 'inches'),
    },
    'ft': {
        'unit': MahaMulticsUnit((0, 1, 0, 0, 0, 0, 0), scale=0.3048, offset=0),
        'tags': ('length',),
        'name': 'foot',
        'aliases': ('foot', 'feet'),
    },

    ## VOLUME ----------------------------------------------------------------
    'L': {
        'unit': MahaMulticsUnit((0, 3, 0, 0, 0, 0, 0), scale=0.001, offset=0),
        'tags': ('volume',),
        'name': 'liter',
        'aliases': ('liter', 'liters'),
    },

    ## MASS ------------------------------------------------------------------
    'g': {
        'unit': MahaMulticsUnit((1, 0, 0, 0, 0, 0, 0), scale=0.001, offset=0),
        'tags': ('mass',),
        'name': 'gram',
        'aliases': ('gram', 'grams'),
    },
    'lbm': {
        'unit': MahaMulticsUnit((1, 0, 0, 0, 0, 0, 0), scale=_LBM_TO_KG_SCALE,
                                offset=0),
        'tags': ('mass',),
        'name': 'pound-mass',
        'description': 'Avoirdupois pound-mass',
    },

    ## FORCE -----------------------------------------------------------------
    'N': {
        'unit': MahaMulticsUnit((1, 1, -2, 0, 0, 0, 0), scale=1, offset=0),
        'tags': ('force',),
        'name': 'Newtons',
        'aliases': ('Newton', 'Newtons'),
    },
    'kN': {
        'unit': MahaMulticsUnit((1, 1, -2, 0, 0, 0, 0), scale=1000, offset=0),
        'tags': ('force',),
        'name': 'kilonewtons',
        'aliases': ('kilonewton', 'kilonewtons'),
    },
    'lbf': {
        'unit': MahaMulticsUnit((1, 1, -2, 0, 0, 0, 0), scale=_LBF_TO_N_SCALE,
                                offset=0),
        'tags': ('force',),
        'name': 'pound-force',
        'description': 'Avoirdupois pound-force',
    },

    ## PRESSURE --------------------------------------------------------------
    'Pa_a': {
        'unit': MahaMulticsUnit((1, -1, -2, 0, 0, 0, 0), scale=1, offset=0),
        'tags': ('pressure',),
        'name': 'pascal-absolute',
        'description': 'Absolute pressure in units of pascals',
    },
    'Pa': {
        'unit': MahaMulticsUnit((1, -1, -2, 0, 0, 0, 0), scale=1,
                                offset=ATMOSPHERIC_PRESSURE_PA),
        'tags': ('pressure',),
        'name': 'pascal-gauge',
        'description': 'Gauge pressure in units of pascals',
    },
    'MPa_a': {
        'unit': MahaMulticsUnit((1, -1, -2, 0, 0, 0, 0), scale=1e6, offset=0),
        'tags': ('pressure',),
        'name': 'megapascal-absolute',
        'description': 'Absolute pressure in units of megapascals',
    },
    'GPa_a': {
        'unit': MahaMulticsUnit((1, -1, -2, 0, 0, 0, 0), scale=1e9, offset=0),
        'tags': ('pressure',),
        'name': 'gigapascal-absolute',
        'description': 'Absolute pressure in units of gigapascals',
    },
    'bar_a': {
        'unit': MahaMulticsUnit((1, -1, -2, 0, 0, 0, 0), scale=1e5, offset=0),
        'tags': ('pressure',),
        'name': 'bar-absolute',
        'description': 'Absolute pressure in units of bar',
    },
    'bar': {
        'unit': MahaMulticsUnit((1, -1, -2, 0, 0, 0, 0), scale=1e5,
                                offset=ATMOSPHERIC_PRESSURE_PA),
        'tags': ('pressure',),
        'name': 'bar-gauge',
        'description': 'Gauge pressure in units of bar',
    },
    'psi_a': {
        'unit': MahaMulticsUnit((1, -1, -2, 0, 0, 0, 0), scale=_PSI_TO_PA_SCALE,
                                offset=0),
        'tags': ('pressure',),
        'name': 'psi-absolute',
        'description': 'Absolute pressure in units of pounds per square inch',
    },
    'psi': {
        'unit': MahaMulticsUnit((1, -1, -2, 0, 0, 0, 0), scale=_PSI_TO_PA_SCALE,
                                offset=ATMOSPHERIC_PRESSURE_PA),
        'tags': ('pressure',),
        'name': 'psi-gauge',
        'description': 'Gauge pressure in units of pounds per square inch',
    },

    ## TEMPERATURE -----------------------------------------------------------
    'degC': {
        'unit': MahaMulticsUnit((0, 0, 0, 1, 0, 0, 0), scale=1,
                                offset=_DEG_C_TO_K_OFFSET),
        'tags': ('temperature',),
        'name': 'degrees-Celsius',
        'description': 'Degrees Celsius',
    },
    'degF': {
        'unit': MahaMulticsUnit((0, 0, 0, 1, 0, 0, 0), scale=_DEG_F_TO_K_SCALE,
                                offset=_DEG_F_TO_K_OFFSET),
        'tags': ('temperature',),
        'name': 'degrees-Fahrenheit',
        'description': 'Degrees Fahrenheit',
    },
    'degR': {
        'unit': MahaMulticsUnit((0, 0, 0, 1, 0, 0, 0), scale=_DEG_F_TO_K_SCALE,
                                offset=0),
        'tags': ('temperature',),
        'name': 'degrees-Rankine',
        'description': 'Degrees Rankine',
    },
    'degC_diff': {
        'unit': MahaMulticsUnit((0, 0, 0, 1, 0, 0, 0), scale=1, offset=0),
        'tags': ('temperature',),
        'name': 'degrees-Celsius-change',
        'description': 'Change in temperature in degrees Celsius',
    },
    'degF_diff': {
        'unit': MahaMulticsUnit((0, 0, 0, 1, 0, 0, 0), scale=_DEG_F_TO_K_SCALE,
                                offset=0),
        'tags': ('temperature',),
        'name': 'degrees-Fahrenheit-change',
        'description': 'Change in temperature in degrees Fahrenheit',
    },

    # ABSOLUTE/DYNAMIC VISCOSITY ---------------------------------------------
    'cP': {
        'unit': MahaMulticsUnit((1, -1, -1, 0, 0, 0, 0), scale=0.001, offset=0),
        'tags': ('viscosity',),
        'name': 'centipoise',
        'description': 'Measure of absolute/dynamic viscosity',
    },

    # KINEMATIC VISCOSITY ----------------------------------------------------
    'St': {
        'unit': MahaMulticsUnit((0, 2, -1, 0, 0, 0, 0), scale=1e-4, offset=0),
        'tags': ('viscosity',),
        'name': 'stoke',
        'description': 'Measure of kinematic viscosity',
    },
    'cSt': {
        'unit': MahaMulticsUnit((0, 2, -1, 0, 0, 0, 0), scale=1e-6, offset=0),
        'tags': ('viscosity',),
        'name': 'centistoke',
        'description': 'Measure of kinematic viscosity',
    },

    # WORK/ENERGY ------------------------------------------------------------
    'J': {
        'unit': MahaMulticsUnit((1, 2, -2, 0, 0, 0, 0), scale=1, offset=0),
        'tags': ('work', 'energy',),
        'name': 'joule',
        'aliases': ('joule', 'joules'),
    },
    'kJ': {
        'unit': MahaMulticsUnit((1, 2, -2, 0, 0, 0, 0), scale=1000, offset=0),
        'tags': ('work', 'energy',),
        'name': 'kilojoule',
        'aliases': ('kilojoule', 'kilojoules'),
    },

    # POWER ------------------------------------------------------------------
    'W': {
        'unit': MahaMulticsUnit((1, 2, -3, 0, 0, 0, 0), scale=1, offset=0),
        'tags': ('power',),
        'name': 'watt',
        'aliases': ('watt', 'watts'),
    },
    'kW': {
        'unit': MahaMulticsUnit((1, 2, -3, 0, 0, 0, 0), scale=1000, offset=0),
        'tags': ('power',),
        'name': 'kilowatt',
        'aliases': ('kilowatt', 'kilowatts'),
    },

    ## TIME ------------------------------------------------------------------
    'ms': {
        'unit': MahaMulticsUnit((0, 0, 1, 0, 0, 0, 0), scale=0.001, offset=0),
        'tags': ('time',),
        'name': 'millisecond',
        'aliases': ('millisecond', 'milliseconds'),
    },
    'min': {
        'unit': MahaMulticsUnit((0, 0, 1, 0, 0, 0, 0), scale=60, offset=0),
        'tags': ('time',),
        'name': 'minute',
        'aliases': ('minute', 'minutes'),
    },
    'hr': {
        'unit': MahaMulticsUnit((0, 0, 1, 0, 0, 0, 0), scale=3600, offset=0),
        'tags': ('time',),
        'name': 'hour',
        'aliases': ('hour', 'hours'),
    },

    ## FREQUENCY -------------------------------------------------------------
    'Hz': {
        'unit': MahaMulticsUnit((0, 0, -1, 0, 0, 0, 0), scale=1, offset=0),
        'tags': ('frequency',),
        'name': 'hertz',
        'aliases': ('hertz',),
    },
    'kHz': {
        'unit': MahaMulticsUnit((0, 0, -1, 0, 0, 0, 0), scale=1e3, offset=0),
        'tags': ('frequency',),
        'name': 'kilohertz',
        'aliases': ('kilohertz',),
    },
    'MHz': {
        'unit': MahaMulticsUnit((0, 0, -1, 0, 0, 0, 0), scale=1e6, offset=0),
        'tags': ('frequency',),
        'name': 'megahertz',
        'aliases': ('megahertz',),
    },
    'GHz': {
        'unit': MahaMulticsUnit((0, 0, -1, 0, 0, 0, 0), scale=1e9, offset=0),
        'tags': ('frequency',),
        'name': 'gigahertz',
        'aliases': ('gigahertz',),
    },
}


class MahaMulticsUnitConverter(pyxx.units.UnitConverter):
    """A unit converter using the SI system of units, pre-filled with
    common units

    This class is identical in function to a :py:class:`UnitConveter`, with
    the exception that it has been initialized with a number of commonly-used
    units.  All units in a :py:class:`UnitConveterSI` are defined in terms of
    the SI system (https://www.nist.gov/pml/owm/metric-si/si-units), with the
    following sequence of seven base units (used in the
    :py:attr:`Unit.base_unit_exps` attribute of all units in the unit
    converter):

    1. Mass: kilogram [:math:`kg`]
    2. Length: meter [:math:`m`]
    3. Time: second [:math:`s`]
    4. Temperature: Kelvin [:math:`K`]
    5. Amount of substance: mole [:math:`mol`]
    6. Electric current: ampere [:math:`A`]
    7. Luminous intensity: candela [:math:`cd`]

    For a list of units available by default, please refer to the
    :ref:`section-unitconverter_units` page.
    """

    def __init__(self):
        """Creates a new unit converter using the SI system of units and
        populated with commonly-used units.

        Creates an new :py:class:`UnitConverterSI` instance.  This unit
        converter uses the SI system (specifically, the system of units
        referenced by :py:class:`UnitSystemSI`) and is pre-filled with a
        set of commonly-used units.
        """
        super().__init__(unit_system=MahaMulticsUnitSystem())

        for key, unit_data in _MAHA_MULTICS_DEFAULT_UNITS.items():
            self.add_unit(key=key, **unit_data)



"""**Classes for parsing Maha Multics input and output files**

The :py:mod:`mahautils.multics` module provides file parsers compatible with
Maha Multics input and output files.  These objects are intended to aid in
pre- and post-processing tasks.
"""

from .configfile import MahaMulticsConfigFile
from .fluidprop import FluidPropertyFile
from .polygonfile import PolygonFile
from .units import (
    MahaMulticsUnit,
    MahaMulticsUnitConverter,
    MahaMulticsUnitSystem,
)
from .simresults import SimResults
from .sim_results_viewer.app import main as run_sim_results_viewer
from .vtk import VTKFile



"""Configuration panel tab for changing plot settings and selecting which
data to plot.
"""

# Mypy type checking disabled for packages that are not PEP 561-compliant
import dash_bootstrap_components as dbc  # type: ignore

from .settings_general import plot_settings_general
from .settings_x import plot_settings_x
from .settings_y import plot_settings_y


def plot_settings_tab():
    """Creates the configuration panel tab (including sub-tabs) where display
    options can be configured"""
    tabs = [
        dbc.Tab(
            plot_settings_general(),
            label='General',
            active_tab_class_name='fw-bold',
        ),
        dbc.Tab(
            plot_settings_x(),
            label='X-Axis',
            active_tab_class_name='fw-bold',
        ),
        dbc.Tab(
            plot_settings_y(),
            label='Y-Axes',
            active_tab_class_name='fw-bold',
        ),
    ]

    return dbc.Tabs(tabs, style={'marginTop': '5px'})



"""Code for generating the header of the simulation results viewer.
"""

# Mypy type checking disabled for packages that are not PEP 561-compliant
import dash                              # type: ignore
import dash_bootstrap_components as dbc  # type: ignore

from .constants import GUI_SHORT_NAME, PROJECT_NAME

button_style = {
    'borderRadius': '8px',
    'marginTop': '5px',
    'marginLeft': '6px',
    'marginRight': '2px',
    'marginBottom': '5px',
}


def app_header():
    """Creates a formatted header for the application"""
    contents = [
        dbc.Stack([
            # Left side of header: app name
            dbc.Row([
                dash.html.H1(
                    f'{PROJECT_NAME} {GUI_SHORT_NAME}',
                    style={
                        'fontSize': 24,
                        'marginLeft': '10px',
                        'marginRight': '10px',
                    },
                ),
            ]),

            # Right side of header: buttons to open plot configuration pane
            # and app information box
            dbc.Row(
                dash.html.Div(
                    [
                        dbc.Button(
                            dash.html.I(className='fa fa-wrench'),
                            id='plot-config-button',
                            size='sm',
                            style=button_style,
                        ),
                        dbc.Tooltip(
                            f'Configure {GUI_SHORT_NAME} settings',
                            target='plot-config-button',
                            trigger='hover',
                        ),
                        dbc.Button(
                            dash.html.I(className='fa fa-question-circle'),
                            id='info-button',
                            size='sm',
                            style=button_style,
                        ),
                        dbc.Tooltip(
                            f'View {GUI_SHORT_NAME} version and acknowledgments',
                            target='info-button',
                            trigger='hover',
                        ),
                        dash.html.A(
                            dbc.Button(
                                dash.html.I(
                                    className='fa fa-up-right-from-square'
                                ),
                                id='new-instance-button',
                                size='sm',
                                style=button_style,
                            ),
                            href='',
                            target='_blank',
                        ),
                        dbc.Tooltip(
                            f'Open a new {GUI_SHORT_NAME} instance',
                            target='new-instance-button',
                            trigger='hover',
                        ),
                    ],
                ),
                style={
                    'marginLeft': 'auto',
                    'marginRight': '0px',
                },
            ),
        ], direction='horizontal'),

        # Horizontal line below title
        dash.html.Hr(),
    ]

    return dbc.Row(children=contents)
